#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Paul Franklin
# Copyright (C) 2023       Oliver Lehmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A plugin to verify the data against user-adjusted tests.
This is the research tool, not the low-level data ingerity check.

Note that this tool has an old heritage (20-Oct-2002 at least) and
so there are vestages of earlier ways of doing things which have not
been converted to a more-modern way.  For instance the way the tool
options are defined (and read in) is not done the way it would be now.
"""

# pylint: disable=not-callable
# pylint: disable=no-self-use

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------

import os
import pickle
import statistics
from hashlib import md5

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GObject

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.errors import WindowActiveError
from gramps.gen.const import URL_MANUAL_PAGE, USER_DATA_VERSION
from gramps.gen.lib import (
    ChildRefType,
    EventRoleType,
    EventType,
    FamilyRelType,
    NameType,
    Person,
)
from gramps.gen.lib.family import Family
from gramps.gen.lib.date import Today
from gramps.gui.editors import EditPerson, EditFamily
from gramps.gen.utils.db import family_name
from gramps.gen.utils.lru import LRU
from gramps.gui.display import display_help
from gramps.gui.managedwindow import ManagedWindow
from gramps.gen.updatecallback import UpdateCallback
from gramps.gui.plug import tool
from gramps.gui.glade import Glade

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Verify_the_Data", "manual")


# -------------------------------------------------------------------------
#
# temp storage and related functions
#
# -------------------------------------------------------------------------
class VerifyFamily:
    def __init__(self, db, family: Family):
        self.handle = ""
        self.marr_date = 0
        self.divo_date = 0
        self.events_in_wrong_order = False
        self.events_of_type_unknown = False
        self.name = ""
        self.mother_handle = ""
        self.father_handle = ""
        self.gramps_id = ""
        self.child_ref_list = {}
        self.relationship = None

        if family is None:
            return

        self.handle = family.get_handle()
        self.name = family_name(family, db)
        self.mother_handle = family.get_mother_handle()
        self.father_handle = family.get_father_handle()
        self.gramps_id = family.get_gramps_id()
        self.child_ref_list = family.get_child_ref_list()
        self.relationship = family.get_relationship()

        prev_date = 0
        for event_ref in family.get_event_ref_list():
            event = db.get_event_from_handle(event_ref.ref)
            date_obj = event.get_date_object()
            if date_obj and date_obj.get_day() != 0 and date_obj.get_month() != 0:
                if prev_date > date_obj.get_sort_value() > 0:
                    self.events_in_wrong_order = True
                prev_date = date_obj.get_sort_value()

            if event_ref.get_role() == EventRoleType.UNKNOWN:
                self.events_of_type_unknown = True
                continue
            if (
                event_ref.get_role() == EventRoleType.FAMILY
                or event_ref.get_role() == EventRoleType.PRIMARY
            ):
                etype = event.get_type()
                date_obj = event.get_date_object()
                if etype == EventType.MARRIAGE:
                    self.marr_date = date_obj.get_sort_value()
                elif etype == EventType.DIVORCE:
                    self.divo_date = date_obj.get_sort_value()

    def get_marriage_date(self):
        return self.marr_date

    def get_divorce_date(self):
        return self.divo_date

    def get_mother_handle(self):
        return self.mother_handle

    def get_father_handle(self):
        return self.father_handle

    def is_events_of_type_unknown(self):
        return self.events_of_type_unknown

    def is_events_in_wrong_order(self):
        return self.events_in_wrong_order

    def get_name(self):
        return self.name

    def get_child_ref_list(self):
        return self.child_ref_list

    def get_relationship(self):
        return self.relationship

    def get_handle(self):
        return self.handle


class VerifyPerson:
    def __init__(self, db, person: Person):
        self.handle = ""
        self.birth_date = [0, 0]
        self.death_date = [0, 0]
        self.bapt_date = [0, 0]
        self.bury_date = [0, 0]
        self.death = False
        self.birth_date_invalid = False
        self.death_date_invalid = False
        self.events_of_type_unknown = False
        self.name = ""
        self.surname = ""
        self.name_type = NameType.UNKNOWN
        self.gramps_id = ""
        self.gender = Person.UNKNOWN
        self.family_handle_list = {}
        self.parent_family_handle_list = {}
        self.events_in_wrong_order = False

        if person is None:
            return

        self.handle = person.get_handle()
        self.name = person.get_primary_name().get_name()
        self.surname = person.get_primary_name().get_surname()
        self.gramps_id = person.get_gramps_id()
        self.gender = person.get_gender()
        self.family_handle_list = person.get_family_handle_list()
        self.parent_family_handle_list = person.get_parent_family_handle_list()
        self.death = bool(person.get_death_ref())
        self.name_type = person.get_primary_name().get_type()

        prev_date = 0
        for event_ref in person.get_event_ref_list():
            event = db.get_event_from_handle(event_ref.ref)
            date_obj = event.get_date_object()
            if date_obj and date_obj.get_day() != 0 and date_obj.get_month() != 0:
                if prev_date > date_obj.get_sort_value() > 0:
                    self.events_in_wrong_order = True
                prev_date = date_obj.get_sort_value()

            if event and event_ref.get_role() == EventRoleType.UNKNOWN:
                self.events_of_type_unknown = True
                continue
            if event and event_ref.get_role() == EventRoleType.PRIMARY:
                etype = event.get_type()
                date_obj = event.get_date_object()
                if date_obj:
                    if date_obj.get_day() == 0 or date_obj.get_month() == 0:
                        exact_date = 0
                    else:
                        exact_date = date_obj.get_sort_value()

                    if etype == EventType.BAPTISM or (
                        etype == EventType.CHRISTEN and self.bapt_date[1] == 0
                    ):
                        self.bapt_date[0] = exact_date
                        self.bapt_date[1] = date_obj.get_sort_value()
                    elif etype == EventType.BURIAL:
                        self.bury_date[0] = exact_date
                        self.bury_date[1] = date_obj.get_sort_value()
                    elif etype == EventType.BIRTH:
                        if not date_obj.get_valid():
                            self.birth_date_invalid = True
                        self.birth_date[0] = exact_date
                        self.birth_date[1] = date_obj.get_sort_value()
                    elif etype == EventType.DEATH:
                        if not date_obj.get_valid():
                            self.death_date_invalid = True
                        self.death_date[0] = exact_date
                        self.death_date[1] = date_obj.get_sort_value()

    def get_birth_date(self, estimate=False):
        return self.birth_date[int(estimate)]

    def get_death_date(self, estimate=False):
        return self.death_date[int(estimate)]

    def get_bapt_date(self, estimate=False):
        return self.bapt_date[int(estimate)]

    def get_bury_date(self, estimate=False):
        return self.bury_date[int(estimate)]

    def get_name(self):
        return self.name

    def get_surname(self):
        return self.surname

    def get_name_type(self):
        return self.name_type

    def get_family_handle_list(self):
        return self.family_handle_list

    def get_gender(self):
        return self.gender

    def get_parent_family_handle_list(self):
        return self.parent_family_handle_list

    def get_death(self):
        return self.death

    def is_birth_date_invalid(self):
        return self.birth_date_invalid

    def is_death_date_invalid(self):
        return self.death_date_invalid

    def is_events_of_type_unknown(self):
        return self.events_of_type_unknown

    def is_events_in_wrong_order(self):
        return self.events_in_wrong_order

    def get_gramps_id(self):
        return self.gramps_id

    def get_handle(self):
        return self.handle


_person_cache_size = 20000
_family_cache_size = 10000
_person_cache = LRU(_person_cache_size + 1)
_family_cache = LRU(_family_cache_size + 1)
_today = Today().get_sort_value()


def find_person(db, handle):
    """find a person, given a handle"""
    if handle not in _person_cache:
        person = db.get_person_from_handle(handle)
        verify_person = VerifyPerson(db, person)
        _person_cache[handle] = verify_person
    return _person_cache[handle]


def find_family(db, handle):
    """find a family, given a handle"""
    if handle not in _family_cache:
        family = db.get_family_from_handle(handle)
        verify_family = VerifyFamily(db, family)
        _family_cache[handle] = verify_family
    return _family_cache[handle]


def preload_person_cache(db):
    """puts all existing people in the cache"""
    for person in db.iter_people():
        verify_person = VerifyPerson(db, person)
        _person_cache[person.get_handle()] = verify_person


def preload_family_cache(db):
    """puts all existing families in the cache"""
    for family in db.iter_families():
        verify_family = VerifyFamily(db, family)
        _family_cache[family.get_handle()] = verify_family


def clear_cache():
    """clear the cache"""
    _person_cache.clear()
    _family_cache.clear()


# -------------------------------------------------------------------------
#
# helper functions
#
# -------------------------------------------------------------------------
def get_age_at_death(person, estimate):
    """get a person's age at death"""
    birth_date = person.get_birth_date(estimate)
    death_date = person.get_death_date(estimate)
    if (birth_date > 0) and (death_date > 0):
        return death_date - birth_date
    return 0


def get_father(db, family):
    """get a family's father"""
    if not family:
        return VerifyPerson(None, None)
    father_handle = family.get_father_handle()
    if father_handle:
        return find_person(db, father_handle)
    return VerifyPerson(None, None)


def get_mother(db, family):
    """get a family's mother"""
    if not family:
        return VerifyPerson(None, None)
    mother_handle = family.get_mother_handle()
    if mother_handle:
        return find_person(db, mother_handle)
    return VerifyPerson(None, None)


def get_child_birth_dates(db, family, estimate):
    """get a family's children's birth dates"""
    dates = []
    for child_ref in family.get_child_ref_list():
        child = find_person(db, child_ref.ref)
        child_birth_date = child.get_birth_date(estimate)
        if child_birth_date > 0:
            dates.append(child_birth_date)
    return dates


def get_n_children(db, person):
    """get the number of a family's children"""
    number = 0
    for family_handle in person.get_family_handle_list():
        family = find_family(db, family_handle)
        if family:
            number += len(family.get_child_ref_list())
    return number


# -------------------------------------------------------------------------
#
# Actual tool
#
# -------------------------------------------------------------------------
class Verify(tool.Tool, ManagedWindow, UpdateCallback):
    """
    A plugin to verify the data against user-adjusted tests.
    This is the research tool, not the low-level data integrity check.
    """

    def __init__(self, dbstate, user, options_class, name, callback=None):
        """initialize things"""
        uistate = user.uistate
        self.label = _("Data Verify tool")
        self.v_r = None
        tool.Tool.__init__(self, dbstate, options_class, name)
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        if uistate:
            UpdateCallback.__init__(self, self.uistate.pulse_progressbar)

        self.dbstate = dbstate
        if uistate:
            self.init_gui()
        else:
            self.add_results = self.add_results_cli
            self.run_the_tool(cli=True)

    def add_results_cli(self, results):
        """print data for the user, no GUI"""
        (msg, gramps_id, name, the_type, rule_id, severity, handle) = results
        severity_str = "S"
        if severity == Rule.WARNING:
            severity_str = "W"
        elif severity == Rule.ERROR:
            severity_str = "E"
        # Translators: needed for French+Arabic, ignore otherwise
        print(
            _("%(severity)s: %(msg)s, %(type)s: %(gid)s, %(name)s")
            % {
                "severity": severity_str,
                "msg": msg,
                "type": the_type,
                "gid": gramps_id,
                "name": name,
            }
        )

    def init_gui(self):
        """Draw dialog and make it handle everything"""
        self.v_r = None
        self.top = Glade()
        self.top.connect_signals(
            {
                "destroy_passed_object": self.close,
                "on_help_clicked": self.on_help_clicked,
                "on_verify_ok_clicked": self.on_apply_clicked,
                "on_delete_event": self.close,
            }
        )

        window = self.top.toplevel
        self.set_window(window, self.top.get_object("title"), self.label)
        self.setup_configs("interface.verify", 650, 400)

        o_dict = self.options.handler.options_dict
        for option in o_dict:
            if option in ["estimate_age", "invdate"]:
                self.top.get_object(option).set_active(o_dict[option])
            else:
                self.top.get_object(option).set_value(o_dict[option])
        self.show()

    def build_menu_names(self, obj):
        """build the menu names"""
        return (_("Tool settings"), self.label)

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def on_apply_clicked(self, obj):
        """event handler for user clicking the OK button: start things"""
        run_button = self.top.get_object("button4")
        close_button = self.top.get_object("button5")
        run_button.set_sensitive(False)
        close_button.set_sensitive(False)
        o_dict = self.options.handler.options_dict
        for option in o_dict:
            if option in ["estimate_age", "invdate"]:
                o_dict[option] = self.top.get_object(option).get_active()
            else:
                o_dict[option] = self.top.get_object(option).get_value_as_int()

        try:
            self.v_r = VerifyResults(
                self.dbstate, self.uistate, self.track, self.top, self.close
            )
            self.add_results = self.v_r.add_results
            self.v_r.load_ignored(self.db.full_name)
        except WindowActiveError:
            pass
        except AttributeError:  # VerifyResults.load_ignored was not run
            self.v_r.ignores = {}

        self.track_ref_for_deletion("v_r")
        self.track_ref_for_deletion("add_results")
        self.track_ref_for_deletion("options")

        self.uistate.set_busy_cursor(True)
        self.uistate.progress.show()
        busy_cursor = Gdk.Cursor.new_for_display(
            Gdk.Display.get_default(), Gdk.CursorType.WATCH
        )
        self.window.get_window().set_cursor(busy_cursor)
        try:
            self.v_r.window.get_window().set_cursor(busy_cursor)
        except AttributeError:
            pass

        self.run_the_tool(cli=False)

        self.uistate.progress.hide()
        self.uistate.set_busy_cursor(False)
        try:
            self.window.get_window().set_cursor(None)
            self.v_r.window.get_window().set_cursor(None)
        except AttributeError:
            pass
        run_button.set_sensitive(True)
        close_button.set_sensitive(True)
        self.reset()

        # Save options
        self.options.handler.save_options()

    def process_family(
        self,
        cli,
        verify_family,
        hwdif,
        yngmar,
        oldmar,
        oldmom,
        olddad,
        yngmom,
        yngdad,
        cbspan,
        cspace,
        estimate_age,
    ):
        """runs all family rules on the given family"""
        rule_list = [
            SameSexFamily(self.db, verify_family),
            FemaleHusband(self.db, verify_family),
            MaleWife(self.db, verify_family),
            SameSurnameFamily(self.db, verify_family),
            LargeAgeGapFamily(self.db, verify_family, hwdif, estimate_age),
            MarriageBeforeBirth(self.db, verify_family, estimate_age),
            MarriageAfterDeath(self.db, verify_family, estimate_age),
            EarlyMarriage(self.db, verify_family, yngmar, estimate_age),
            LateMarriage(self.db, verify_family, oldmar, estimate_age),
            OldParent(self.db, verify_family, oldmom, olddad, estimate_age),
            YoungParent(self.db, verify_family, yngmom, yngdad, estimate_age),
            UnbornParent(self.db, verify_family, estimate_age),
            DeadParent(self.db, verify_family, estimate_age),
            LargeChildrenSpan(self.db, verify_family, cbspan, estimate_age),
            LargeChildrenAgeDiff(self.db, verify_family, cspace, estimate_age),
            MarriedRelation(self.db, verify_family),
            ChildrenOrderIncorrect(self.db, verify_family, estimate_age),
            FamilyHasEventsOfTypeUnknown(self.db, verify_family),
            FamilyHasEventsInWrongOrder(self.db, verify_family),
        ]

        for rule in rule_list:
            if rule.broken():
                self.add_results(rule.report_itself())

        if not cli:
            self.update()

    def process_person(
        self,
        cli,
        verify_person,
        oldage,
        wedder,
        oldunm,
        mxchilddad,
        mxchildmom,
        invdate,
        estimate_age,
    ):
        """runs all person rules on the given person"""
        rule_list = [
            BirthAfterBapt(self.db, verify_person),
            DeathBeforeBapt(self.db, verify_person),
            BirthAfterBury(self.db, verify_person),
            DeathAfterBury(self.db, verify_person),
            BirthAfterDeath(self.db, verify_person),
            BaptAfterBury(self.db, verify_person),
            OldAge(self.db, verify_person, oldage, estimate_age),
            OldAgeButNoDeath(self.db, verify_person, oldage, estimate_age),
            UnknownGender(self.db, verify_person),
            MultipleParents(self.db, verify_person),
            MarriedOften(self.db, verify_person, wedder),
            OldUnmarried(self.db, verify_person, oldunm, estimate_age),
            TooManyChildren(self.db, verify_person, mxchilddad, mxchildmom),
            Disconnected(self.db, verify_person),
            InvalidBirthDate(self.db, verify_person, invdate),
            InvalidDeathDate(self.db, verify_person, invdate),
            BirthEqualsDeath(self.db, verify_person),
            BirthEqualsMarriage(self.db, verify_person),
            DeathEqualsMarriage(self.db, verify_person),
            BaptTooLate(self.db, verify_person),
            BuryTooLate(self.db, verify_person),
            FamilyOrderIncorrect(self.db, verify_person, estimate_age),
            PersonHasEventsOfTypeUnknown(self.db, verify_person),
            PersonHasEventsInWrongOrder(self.db, verify_person),
        ]
        for rule in rule_list:
            if rule.broken():
                self.add_results(rule.report_itself())

        if not cli:
            self.update()

    def run_the_tool(self, cli=False):
        """run the tool"""

        n_people = self.db.get_number_of_people()
        n_families = self.db.get_number_of_families()
        if n_people <= _person_cache_size:
            preload_person_cache(self.db)
        if n_families <= _family_cache_size:
            preload_family_cache(self.db)

        options_dict = self.options.handler.options_dict

        if self.v_r:
            self.v_r.real_model.clear()

        self.set_total(n_people + n_families)

        family_handles = set(self.db.get_family_handles())

        for person in self.db.iter_people():
            verify_person = VerifyPerson(self.db, person)
            _person_cache[verify_person.get_handle()] = verify_person
            for family_handle in verify_person.get_family_handle_list():
                if family_handle in family_handles:
                    verify_family = find_family(self.db, family_handle)
                    self.process_family(
                        cli,
                        verify_family,
                        options_dict["hwdif"],
                        options_dict["yngmar"],
                        options_dict["oldmar"],
                        options_dict["oldmom"],
                        options_dict["olddad"],
                        options_dict["yngmom"],
                        options_dict["yngdad"],
                        options_dict["cbspan"],
                        options_dict["cspace"],
                        options_dict["estimate_age"],
                    )
                    family_handles.remove(family_handle)

            self.process_person(
                cli,
                verify_person,
                options_dict["oldage"],
                options_dict["wedder"],
                options_dict["oldunm"],
                options_dict["mxchilddad"],
                options_dict["mxchildmom"],
                options_dict["invdate"],
                options_dict["estimate_age"],
            )

        # Family-based rules - for any left over families (families without any spouses)
        for family_handle in family_handles:
            verify_family = find_family(self.db, family_handle)

            self.process_family(
                cli,
                verify_family,
                options_dict["hwdif"],
                options_dict["yngmar"],
                options_dict["oldmar"],
                options_dict["oldmom"],
                options_dict["olddad"],
                options_dict["yngmom"],
                options_dict["yngdad"],
                options_dict["cbspan"],
                options_dict["cspace"],
                options_dict["estimate_age"],
            )

        clear_cache()
        self.update = None  # Needed for garbage collection


# -------------------------------------------------------------------------
#
# Display the results
#
# -------------------------------------------------------------------------
class VerifyResults(ManagedWindow):
    """GUI class to show the results in another dialog"""

    IGNORE_COL = 0
    WARNING_COL = 1
    OBJ_ID_COL = 2
    OBJ_NAME_COL = 3
    OBJ_TYPE_COL = 4
    RULE_ID_COL = 5
    OBJ_HANDLE_COL = 6
    FG_COLOR_COL = 7
    TRUE_COL = 8
    SHOW_COL = 9

    def __init__(self, dbstate, uistate, track, glade, closeall):
        """initialize things"""
        self.title = _("Data Verification Results")

        ManagedWindow.__init__(self, uistate, track, self.__class__)

        self.dbstate = dbstate
        self.closeall = closeall
        self._set_filename()
        self.top = glade
        window = self.top.get_object("verify_result")
        self.set_window(window, self.top.get_object("title2"), self.title)
        self.setup_configs("interface.verifyresults", 500, 300)
        window.connect("close", self.close)
        close_btn = self.top.get_object("closebutton1")
        close_btn.connect("clicked", self.close)

        self.warn_tree = self.top.get_object("warn_tree")
        self.warn_tree.connect("button_press_event", self.double_click)

        self.selection = self.warn_tree.get_selection()

        self.hide_button = self.top.get_object("hide_button")
        self.hide_button.connect("toggled", self.hide_toggled)

        self.mark_button = self.top.get_object("mark_all")
        self.mark_button.connect("clicked", self.mark_clicked)

        self.unmark_button = self.top.get_object("unmark_all")
        self.unmark_button.connect("clicked", self.unmark_clicked)

        self.invert_button = self.top.get_object("invert_all")
        self.invert_button.connect("clicked", self.invert_clicked)

        self.parent_iter_cache = {}

        self.real_model = Gtk.TreeStore(
            GObject.TYPE_BOOLEAN,
            GObject.TYPE_STRING,
            GObject.TYPE_STRING,
            GObject.TYPE_STRING,
            GObject.TYPE_STRING,
            object,
            GObject.TYPE_STRING,
            GObject.TYPE_STRING,
            GObject.TYPE_BOOLEAN,
            GObject.TYPE_BOOLEAN,
        )
        self.filt_model = self.real_model.filter_new()
        self.filt_model.set_visible_column(VerifyResults.TRUE_COL)
        if hasattr(self.filt_model, "sort_new_with_model"):
            self.sort_model = self.filt_model.sort_new_with_model()
        else:
            self.sort_model = Gtk.TreeModelSort.new_with_model(self.filt_model)
        self.warn_tree.set_model(self.sort_model)

        self.renderer = Gtk.CellRendererText()
        self.img_renderer = Gtk.CellRendererPixbuf()
        self.bool_renderer = Gtk.CellRendererToggle()
        self.bool_renderer.connect("toggled", self.selection_toggled)
        self.track_ref_for_deletion("bool_renderer")

        # Add ignore column
        ignore_column = Gtk.TreeViewColumn(
            _("Mark"), self.bool_renderer, active=VerifyResults.IGNORE_COL
        )
        ignore_column.set_sort_column_id(VerifyResults.IGNORE_COL)
        self.warn_tree.append_column(ignore_column)

        # Add image column
        img_column = Gtk.TreeViewColumn(None, self.img_renderer)
        img_column.set_cell_data_func(self.img_renderer, self.get_image)
        self.warn_tree.append_column(img_column)

        # Add column with the warning text
        warn_column = Gtk.TreeViewColumn(
            _("Warning"),
            self.renderer,
            text=VerifyResults.WARNING_COL,
            foreground=VerifyResults.FG_COLOR_COL,
        )
        warn_column.set_sort_column_id(VerifyResults.WARNING_COL)
        self.warn_tree.append_column(warn_column)

        # Add column with object gramps_id
        id_column = Gtk.TreeViewColumn(
            _("ID"),
            self.renderer,
            text=VerifyResults.OBJ_ID_COL,
            foreground=VerifyResults.FG_COLOR_COL,
        )
        id_column.set_sort_column_id(VerifyResults.OBJ_ID_COL)
        self.warn_tree.append_column(id_column)

        # Add column with object name
        name_column = Gtk.TreeViewColumn(
            _("Name"),
            self.renderer,
            text=VerifyResults.OBJ_NAME_COL,
            foreground=VerifyResults.FG_COLOR_COL,
        )
        name_column.set_sort_column_id(VerifyResults.OBJ_NAME_COL)
        self.warn_tree.append_column(name_column)

        self.show()
        self.window_shown = False

    def _set_filename(self):
        """set the file where people who will be ignored will be kept"""
        db_filename = self.dbstate.db.get_save_path()
        if isinstance(db_filename, str):
            db_filename = db_filename.encode("utf-8")
        md5sum = md5(db_filename)
        self.ignores_filename = os.path.join(
            USER_DATA_VERSION, md5sum.hexdigest() + os.path.extsep + "vfm"
        )

    def load_ignored(self, db_filename):
        """get ready to load the file with the previously-ignored people"""
        ## a new Gramps major version means recreating the .vfm file.
        ## User can copy over old one, with name of new one, but no guarantee
        ## that will work.
        if not self._load_ignored(self.ignores_filename):
            self.ignores = {}

    def _load_ignored(self, filename):
        """load the file with the people who were previously ignored"""
        try:
            try:
                file = open(filename, "rb")
            except IOError:
                return False
            self.ignores = pickle.load(file)
            file.close()
            return True
        except (IOError, EOFError):
            file.close()
            return False

    def save_ignored(self, new_ignores):
        """get ready to save the file with the ignored people"""
        self.ignores = new_ignores
        self._save_ignored(self.ignores_filename)

    def _save_ignored(self, filename):
        """save the file with the people the user wants to ignore"""
        try:
            with open(filename, "wb") as file:
                pickle.dump(self.ignores, file, 1)
            return True
        except IOError:
            return False

    def get_marking(self, handle, rule_id):
        if handle in self.ignores:
            return rule_id in self.ignores[handle]
        return False

    def get_new_marking(self):
        new_ignores = {}
        for row_num in range(len(self.real_model)):
            path = (row_num,)
            row = self.real_model[path]
            ignore = row[VerifyResults.IGNORE_COL]
            if ignore:
                handle = row[VerifyResults.OBJ_HANDLE_COL]
                rule_id = row[VerifyResults.RULE_ID_COL]
                if handle not in new_ignores:
                    new_ignores[handle] = set()
                new_ignores[handle].add(rule_id)
        return new_ignores

    def close(self, *obj):
        """close the dialog and write out the file"""
        new_ignores = self.get_new_marking()
        self.save_ignored(new_ignores)

        ManagedWindow.close(self, *obj)
        self.closeall()

    def set_parent_text(self, parent_iter, msg, num):
        """update the parent rows message including the number of the children"""
        parent_msg = msg + " (" + str(num) + ")"
        self.real_model.set_value(parent_iter, VerifyResults.WARNING_COL, parent_msg)

    def refresh_all_parent_texts(self):
        """sync all parent texts with the number of currently displayed children"""
        parent_iter = self.filt_model.get_iter_first()
        while parent_iter:
            child_iter = self.filt_model.iter_children(parent_iter)
            if child_iter:
                msg = self.filt_model.get_value(child_iter, VerifyResults.WARNING_COL)
                num = self.filt_model.iter_n_children(parent_iter)
                real_parent_iter = self.filt_model.convert_iter_to_child_iter(
                    parent_iter
                )
                self.set_parent_text(real_parent_iter, msg, num)
            parent_iter = self.filt_model.iter_next(parent_iter)

    def hide_toggled(self, button):
        """either hide the marked rows or show all rows"""
        # memorize all currently expanded rows
        expanded_paths = []
        parent_iter = self.sort_model.get_iter_first()
        while parent_iter:
            sort_path = self.sort_model.get_path(parent_iter)
            if self.warn_tree.row_expanded(sort_path):
                filt_path = self.sort_model.convert_path_to_child_path(sort_path)
                expanded_paths.append(
                    self.filt_model.convert_path_to_child_path(filt_path)
                )
            parent_iter = self.sort_model.iter_next(parent_iter)

        self.filt_model = self.real_model.filter_new()
        if button.get_active():
            button.set_label(_("_Show all"))
            self.filt_model.set_visible_column(VerifyResults.SHOW_COL)
        else:
            button.set_label(_("_Hide marked"))
            self.filt_model.set_visible_column(VerifyResults.TRUE_COL)
        if hasattr(self.filt_model, "sort_new_with_model"):
            self.sort_model = self.filt_model.sort_new_with_model()
        else:
            self.sort_model = Gtk.TreeModelSort.new_with_model(self.filt_model)

        self.refresh_all_parent_texts()

        self.warn_tree.set_model(self.sort_model)

        # expand all not filtered rows which already where expanded
        for real_path in expanded_paths:
            filt_path = self.filt_model.convert_child_path_to_path(real_path)
            if filt_path is not None:
                self.warn_tree.expand_row(filt_path, False)

    def set_row_selection(self, row_iter, value):
        """toggle the given rows checkbox"""
        path = self.real_model.get_path(row_iter)
        self.real_model.set_value(row_iter, VerifyResults.IGNORE_COL, value)
        self.real_model.set_value(row_iter, VerifyResults.SHOW_COL, not value)
        self.real_model.row_changed(path, row_iter)

    def selection_toggled(self, cell, path_string):
        """the rows checkbox click handler"""
        sort_path = tuple(map(int, path_string.split(":")))
        filt_path = self.sort_model.convert_path_to_child_path(Gtk.TreePath(sort_path))
        real_path = self.filt_model.convert_path_to_child_path(filt_path)
        row = self.real_model[real_path]
        the_type = row[VerifyResults.OBJ_TYPE_COL]

        ignore = not row[VerifyResults.IGNORE_COL]
        if the_type == Rule.TYPE_GROUP:
            # (un)select all children when the parent gets activly (un)selected
            child_iter = self.real_model.iter_children(row.iter)
            while child_iter:
                self.set_row_selection(child_iter, ignore)
                child_iter = self.real_model.iter_next(child_iter)
        else:
            parent_iter = self.real_model.iter_parent(row.iter)
            parent_ignore = self.real_model.get_value(
                parent_iter, VerifyResults.IGNORE_COL
            )
            if parent_ignore and not ignore:
                # remove the parents selection when a child becomes no longer selected
                self.set_row_selection(parent_iter, False)
            else:
                all_ignored = True
                child_iter = self.real_model.iter_children(parent_iter)
                while child_iter:
                    # check if all children are selected (or not)
                    if self.real_model.get_path(child_iter) == real_path:
                        # the value of the triggering row is not yet synced into
                        #   the model so we can't read it from there
                        value = ignore
                    else:
                        value = self.real_model.get_value(
                            child_iter, VerifyResults.IGNORE_COL
                        )
                    if not value:
                        all_ignored = False
                        break
                    child_iter = self.real_model.iter_next(child_iter)
                if all_ignored:
                    # select parent when all children become selected
                    self.set_row_selection(parent_iter, True)
                elif self.hide_button.get_active():
                    # update parents warning text when view is in filter mode
                    filt_iter = self.filt_model.get_iter(filt_path)
                    filt_parent_iter = self.filt_model.iter_parent(filt_iter)
                    num_children = self.filt_model.iter_n_children(filt_parent_iter)
                    msg = row[VerifyResults.WARNING_COL]
                    self.set_parent_text(parent_iter, msg, num_children - 1)
        self.set_row_selection(row.iter, ignore)

    def mark_unmark(self, mark):
        """either selects or unselects all rows"""
        parent_iter = self.real_model.get_iter_first()
        while parent_iter:
            ignore = self.real_model.get_value(parent_iter, VerifyResults.IGNORE_COL)
            if (mark and not ignore) or not mark:
                # if the parent should be selected but is already selected skip it
                # if the selection should be removed we must always loop through
                #   all children
                child_iter = self.real_model.iter_children(parent_iter)
                while child_iter:
                    self.set_row_selection(child_iter, mark)
                    child_iter = self.real_model.iter_next(child_iter)
                self.set_row_selection(parent_iter, mark)
            parent_iter = self.real_model.iter_next(parent_iter)
        self.filt_model.refilter()
        self.refresh_all_parent_texts()

    def mark_clicked(self, mark_button):
        """the mark button click handler"""
        self.mark_unmark(True)

    def unmark_clicked(self, unmark_button):
        """the unmark button click handler"""
        self.mark_unmark(False)

    def invert_clicked(self, invert_button):
        """invert the current selection"""
        parent_iter = self.real_model.get_iter_first()
        while parent_iter:
            child_iter = self.real_model.iter_children(parent_iter)
            all_ignored = True
            while child_iter:
                ignore = not self.real_model.get_value(
                    child_iter, VerifyResults.IGNORE_COL
                )
                if not ignore:
                    all_ignored = False
                self.set_row_selection(child_iter, ignore)
                child_iter = self.real_model.iter_next(child_iter)
            self.set_row_selection(parent_iter, all_ignored)
            parent_iter = self.real_model.iter_next(parent_iter)
        self.filt_model.refilter()
        self.refresh_all_parent_texts()

    def double_click(self, obj, event):
        """the user wants to edit the selected person or family"""
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            (model, node) = self.selection.get_selected()
            if not node:
                return
            sort_path = self.sort_model.get_path(node)
            filt_path = self.sort_model.convert_path_to_child_path(sort_path)
            real_path = self.filt_model.convert_path_to_child_path(filt_path)
            row = self.real_model[real_path]
            the_type = row[VerifyResults.OBJ_TYPE_COL]
            handle = row[VerifyResults.OBJ_HANDLE_COL]
            if the_type == Rule.TYPE_PERSON:
                try:
                    person = self.dbstate.db.get_person_from_handle(handle)
                    EditPerson(self.dbstate, self.uistate, self.track, person)
                except WindowActiveError:
                    pass
            elif the_type == Rule.TYPE_FAMILY:
                try:
                    family = self.dbstate.db.get_family_from_handle(handle)
                    EditFamily(self.dbstate, self.uistate, self.track, family)
                except WindowActiveError:
                    pass
            elif the_type == Rule.TYPE_GROUP:
                if self.warn_tree.row_expanded(filt_path):
                    self.warn_tree.collapse_row(filt_path)
                else:
                    self.warn_tree.expand_row(filt_path, False)

    def get_image(self, column, cell, model, iter_, user_data=None):
        """flag whether each line is a person or family"""
        the_type = model.get_value(iter_, VerifyResults.OBJ_TYPE_COL)
        if the_type == Rule.TYPE_PERSON:
            cell.set_property("icon-name", "gramps-person")
        elif the_type == Rule.TYPE_FAMILY:
            cell.set_property("icon-name", "gramps-family")
        else:
            cell.set_property("icon-name", None)

    def add_results(self, results):
        """adds the negative result of an evaluated Rule to the model"""
        (msg, gramps_id, name, the_type, rule_id, severity, handle) = results
        ignore = self.get_marking(handle, rule_id)
        if severity == Rule.ERROR:
            line_color = "red"
        else:
            line_color = None

        parent_iter = None
        # rule_id can't be used because there are rules with dynamic messages
        if msg in self.parent_iter_cache:
            parent_iter = self.parent_iter_cache[msg]
        else:
            parent_iter = self.real_model.append(
                None,
                row=[
                    None,
                    msg,
                    None,
                    None,
                    Rule.TYPE_GROUP,
                    rule_id,
                    None,
                    line_color,
                    True,
                    True,
                ],
            )
            self.parent_iter_cache[msg] = parent_iter

        self.real_model.append(
            parent_iter,
            row=[
                ignore,
                msg,
                gramps_id,
                name,
                the_type,
                rule_id,
                handle,
                line_color,
                True,
                not ignore,
            ],
        )

        num = self.real_model.iter_n_children(parent_iter)
        self.set_parent_text(parent_iter, msg, num)

        if not self.window_shown:
            self.window.show()
            self.window_shown = True

    def build_menu_names(self, obj):
        """build the menu names"""
        return (self.title, self.title)


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class VerifyOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        """initialize the options"""
        tool.ToolOptions.__init__(self, name, person_id)

        # Options specific for this report
        self.options_dict = {
            "oldage": 90,
            "hwdif": 30,
            "cspace": 8,
            "cbspan": 25,
            "yngmar": 17,
            "oldmar": 50,
            "oldmom": 48,
            "yngmom": 17,
            "yngdad": 18,
            "olddad": 65,
            "wedder": 3,
            "mxchildmom": 12,
            "mxchilddad": 15,
            "lngwdw": 30,
            "oldunm": 99,
            "estimate_age": 0,
            "invdate": 1,
        }
        # TODO these strings are defined in the glade file (more or less, since
        # those have accelerators), and so are not translated here, but that
        # means that a CLI user who runs gramps in a non-English language and
        # says (for instance) "show=oldage" will see "Maximum age" in English
        # (but I think such a CLI use is very unlikely and so is low priority,
        # especially since the tool's normal CLI output will be translated)
        self.options_help = {
            "oldage": ("=num", "Maximum age", "Age in years"),
            "hwdif": (
                "=num",
                "Maximum husband-wife age difference",
                "Age difference in years",
            ),
            "cspace": (
                "=num",
                "Maximum number of years between children",
                "Number of years",
            ),
            "cbspan": (
                "=num",
                "Maximum span of years for all children",
                "Span in years",
            ),
            "yngmar": ("=num", "Minimum age to marry", "Age in years"),
            "oldmar": ("=num", "Maximum age to marry", "Age in years"),
            "oldmom": ("=num", "Maximum age to bear a child", "Age in years"),
            "yngmom": ("=num", "Minimum age to bear a child", "Age in years"),
            "yngdad": ("=num", "Minimum age to father a child", "Age in years"),
            "olddad": ("=num", "Maximum age to father a child", "Age in years"),
            "wedder": (
                "=num",
                "Maximum number of spouses for a person",
                "Number of spouses",
            ),
            "mxchildmom": (
                "=num",
                "Maximum number of children for a woman",
                "Number of children",
            ),
            "mxchilddad": (
                "=num",
                "Maximum  number of children for a man",
                "Number of chidlren",
            ),
            "lngwdw": (
                "=num",
                "Maximum number of consecutive years "
                "of widowhood before next marriage",
                "Number of years",
            ),
            "oldunm": ("=num", "Maximum age for an unmarried person" "Number of years"),
            "estimate_age": (
                "=0/1",
                "Whether to estimate missing or inexact dates",
                ["Do not estimate", "Estimate dates"],
                True,
            ),
            "invdate": (
                "=0/1",
                "Whether to check for invalid dates" "Do not identify invalid dates",
                "Identify invalid dates",
                True,
            ),
        }


# -------------------------------------------------------------------------
#
# Base classes for different tests -- the rules
#
# -------------------------------------------------------------------------
class Rule:
    """
    Basic class for use in this tool.

    Other rules must inherit from this.
    """

    ID = 0
    TYPE = ""

    ERROR = 1
    WARNING = 2

    SEVERITY = WARNING

    TYPE_PERSON = "Person"
    TYPE_FAMILY = "Family"
    TYPE_GROUP = "Group"

    def __init__(self, db, obj):
        """initialize the rule"""
        self.db = db
        self.obj = obj

    def broken(self):
        """
        Return boolean indicating whether this rule is violated.
        """
        return False

    def get_message(self):
        """return the rule's error message"""
        assert False, "Need to be overriden in the derived class"

    def get_name(self):
        """return the persons or families primary name"""
        return self.obj.get_name()

    def get_handle(self):
        """return the object's handle"""
        return self.obj.get_handle()

    def get_id(self):
        """return the object's gramps_id"""
        return self.obj.gramps_id

    def get_rule_id(self):
        """return the rule's identification number, and parameters"""
        params = self._get_params()
        return (self.ID, params)

    def _get_params(self):
        """return the rule's parameters"""
        return tuple()

    def report_itself(self):
        """return the details about a rule"""
        handle = self.get_handle()
        the_type = self.TYPE
        rule_id = self.get_rule_id()
        severity = self.SEVERITY
        name = self.get_name()
        gramps_id = self.get_id()
        msg = self.get_message()
        self.get_message = None  # Needed for garbage collection
        return (msg, gramps_id, name, the_type, rule_id, severity, handle)


class PersonRule(Rule):
    """
    Person-based class.
    """

    TYPE = Rule.TYPE_PERSON


class FamilyRule(Rule):
    """
    Family-based class.
    """

    TYPE = Rule.TYPE_FAMILY


# -------------------------------------------------------------------------
#
# Actual rules for testing
#
# -------------------------------------------------------------------------
class BirthAfterBapt(PersonRule):
    """test if a person was baptised before their birth"""

    ID = 1
    SEVERITY = Rule.ERROR

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        birth_date = self.obj.get_birth_date()
        bapt_date = self.obj.get_bapt_date()
        birth_ok = birth_date > 0 if birth_date is not None else False
        bapt_ok = bapt_date > 0 if bapt_date is not None else False
        return birth_ok and bapt_ok and birth_date > bapt_date

    def get_message(self):
        """return the rule's error message"""
        return _("Baptism before birth")


class DeathBeforeBapt(PersonRule):
    """test if a person died before their baptism"""

    ID = 2
    SEVERITY = Rule.ERROR

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        death_date = self.obj.get_death_date()
        bapt_date = self.obj.get_bapt_date()
        bapt_ok = bapt_date > 0 if bapt_date is not None else False
        death_ok = death_date > 0 if death_date is not None else False
        return death_ok and bapt_ok and bapt_date > death_date

    def get_message(self):
        """return the rule's error message"""
        return _("Death before baptism")


class BirthAfterBury(PersonRule):
    """test if a person was buried before their birth"""

    ID = 3
    SEVERITY = Rule.ERROR

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        birth_date = self.obj.get_birth_date()
        bury_date = self.obj.get_bury_date()
        birth_ok = birth_date > 0 if birth_date is not None else False
        bury_ok = bury_date > 0 if bury_date is not None else False
        return birth_ok and bury_ok and birth_date > bury_date

    def get_message(self):
        """return the rule's error message"""
        return _("Burial before birth")


class DeathAfterBury(PersonRule):
    """test if a person was buried before their death"""

    ID = 4
    SEVERITY = Rule.ERROR

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        death_date = self.obj.get_death_date()
        bury_date = self.obj.get_bury_date()
        death_ok = death_date > 0 if death_date is not None else False
        bury_ok = bury_date > 0 if bury_date is not None else False
        return death_ok and bury_ok and death_date > bury_date

    def get_message(self):
        """return the rule's error message"""
        return _("Burial before death")


class BirthAfterDeath(PersonRule):
    """test if a person died before their birth"""

    ID = 5
    SEVERITY = Rule.ERROR

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        birth_date = self.obj.get_birth_date()
        death_date = self.obj.get_death_date()
        birth_ok = birth_date > 0 if birth_date is not None else False
        death_ok = death_date > 0 if death_date is not None else False
        return birth_ok and death_ok and birth_date > death_date

    def get_message(self):
        """return the rule's error message"""
        return _("Death before birth")


class BaptAfterBury(PersonRule):
    """test if a person was buried before their baptism"""

    ID = 6
    SEVERITY = Rule.ERROR

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        bapt_date = self.obj.get_bapt_date()
        bury_date = self.obj.get_bury_date()
        bapt_ok = bapt_date > 0 if bapt_date is not None else False
        bury_ok = bury_date > 0 if bury_date is not None else False
        return bapt_ok and bury_ok and bapt_date > bury_date

    def get_message(self):
        """return the rule's error message"""
        return _("Burial before baptism")


class OldAge(PersonRule):
    """test if a person died beyond the age the user has set"""

    ID = 7
    SEVERITY = Rule.WARNING

    def __init__(self, db, person, old_age, est):
        """initialize the rule"""
        PersonRule.__init__(self, db, person)
        self.old_age = old_age
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.old_age, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        age_at_death = get_age_at_death(self.obj, self.est)
        return age_at_death / 365 > self.old_age

    def get_message(self):
        """return the rule's error message"""
        return _("Old age at death")


class UnknownGender(PersonRule):
    """test if a person is neither a male nor a female"""

    ID = 8
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        return self.obj.get_gender() == Person.UNKNOWN

    def get_message(self):
        """return the rule's error message"""
        return _("Unknown gender")


class MultipleParents(PersonRule):
    """test if a person belongs to multiple families"""

    ID = 9
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        n_parent_sets = len(self.obj.get_parent_family_handle_list())
        return n_parent_sets > 1

    def get_message(self):
        """return the rule's error message"""
        return _("Multiple parents")


class MarriedOften(PersonRule):
    """test if a person was married 'often'"""

    ID = 10
    SEVERITY = Rule.WARNING

    def __init__(self, db, person, wedder):
        """initialize the rule"""
        PersonRule.__init__(self, db, person)
        self.wedder = wedder

    def _get_params(self):
        """return the rule's parameters"""
        return (self.wedder,)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        n_spouses = len(self.obj.get_family_handle_list())
        return n_spouses > self.wedder

    def get_message(self):
        """return the rule's error message"""
        return _("Married often")


class OldUnmarried(PersonRule):
    """test if a person was married when they died"""

    ID = 11
    SEVERITY = Rule.WARNING

    def __init__(self, db, person, old_unm, est):
        """initialize the rule"""
        PersonRule.__init__(self, db, person)
        self.old_unm = old_unm
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.old_unm, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        age_at_death = get_age_at_death(self.obj, self.est)
        n_spouses = len(self.obj.get_family_handle_list())
        return age_at_death / 365 > self.old_unm and n_spouses == 0

    def get_message(self):
        """return the rule's error message"""
        return _("Old and unmarried")


class TooManyChildren(PersonRule):
    """test if a person had 'too many' children"""

    ID = 12
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, mx_child_dad, mx_child_mom):
        """initialize the rule"""
        PersonRule.__init__(self, db, obj)
        self.mx_child_dad = mx_child_dad
        self.mx_child_mom = mx_child_mom

    def _get_params(self):
        """return the rule's parameters"""
        return (self.mx_child_dad, self.mx_child_mom)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        n_child = get_n_children(self.db, self.obj)

        if self.obj.get_gender == Person.MALE and n_child > self.mx_child_dad:
            return True

        if self.obj.get_gender == Person.FEMALE and n_child > self.mx_child_mom:
            return True

        return False

    def get_message(self):
        """return the rule's error message"""
        return _("Too many children")


class SameSexFamily(FamilyRule):
    """test if a family's parents are both male or both female"""

    ID = 13
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        same_sex = mother and father and (mother.get_gender() == father.get_gender())
        unknown_sex = mother and (mother.get_gender() == Person.UNKNOWN)
        return same_sex and not unknown_sex

    def get_message(self):
        """return the rule's error message"""
        return _("Same sex marriage")


class FemaleHusband(FamilyRule):
    """test if a family's 'husband' is female"""

    ID = 14
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        father = get_father(self.db, self.obj)
        return father and (father.get_gender() == Person.FEMALE)

    def get_message(self):
        """return the rule's error message"""
        return _("Female husband")


class MaleWife(FamilyRule):
    """test if a family's 'wife' is male"""

    ID = 15
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        return mother and (mother.get_gender() == Person.MALE)

    def get_message(self):
        """return the rule's error message"""
        return _("Male wife")


class SameSurnameFamily(FamilyRule):
    """test if a family's parents were born with the same surname"""

    ID = 16
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        _broken = False

        # Make sure both mother and father exist.
        if mother and father:
            mname = mother.get_surname()
            fname = father.get_surname()
            # Only compare birth names (not married names).
            if (
                mother.get_name_type() == NameType.BIRTH
                and father.get_name_type() == NameType.BIRTH
            ):
                # Empty names don't count.
                if len(mname) != 0 and len(fname) != 0:
                    # Finally, check if the names are the same.
                    if mname == fname:
                        _broken = True

        return _broken

    def get_message(self):
        """return the rule's error message"""
        return _("Husband and wife with the same surname")


class LargeAgeGapFamily(FamilyRule):
    """test if a family's parents were born far apart"""

    ID = 17
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, hw_diff, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.hw_diff = hw_diff
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.hw_diff, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_birth_date = mother.get_birth_date(self.est)
        father_birth_date = father.get_birth_date(self.est)
        mother_birth_date_ok = mother_birth_date > 0
        father_birth_date_ok = father_birth_date > 0
        large_diff = abs(father_birth_date - mother_birth_date) / 365 > self.hw_diff
        return mother_birth_date_ok and father_birth_date_ok and large_diff

    def get_message(self):
        """return the rule's error message"""
        return _("Large age difference between spouses")


class MarriageBeforeBirth(FamilyRule):
    """test if each family's parent was born before the marriage"""

    ID = 18
    SEVERITY = Rule.ERROR

    def __init__(self, db, obj, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.est,)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        marr_date = self.obj.get_marriage_date()
        marr_date_ok = marr_date > 0

        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_birth_date = mother.get_birth_date(self.est)
        father_birth_date = father.get_birth_date(self.est)
        mother_birth_date_ok = mother_birth_date > 0
        father_birth_date_ok = father_birth_date > 0

        father_broken = (
            father_birth_date_ok and marr_date_ok and (father_birth_date > marr_date)
        )
        mother_broken = (
            mother_birth_date_ok and marr_date_ok and (mother_birth_date > marr_date)
        )

        return father_broken or mother_broken

    def get_message(self):
        """return the rule's error message"""
        return _("Marriage before birth")


class MarriageAfterDeath(FamilyRule):
    """test if each family's parent died before the marriage"""

    ID = 19
    SEVERITY = Rule.ERROR

    def __init__(self, db, obj, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.est,)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        marr_date = self.obj.get_marriage_date()
        marr_date_ok = marr_date > 0

        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_death_date = mother.get_death_date(self.est)
        father_death_date = father.get_death_date(self.est)
        mother_death_date_ok = mother_death_date > 0
        father_death_date_ok = father_death_date > 0

        father_broken = (
            father_death_date_ok and marr_date_ok and (father_death_date < marr_date)
        )
        mother_broken = (
            mother_death_date_ok and marr_date_ok and (mother_death_date < marr_date)
        )

        return father_broken or mother_broken

    def get_message(self):
        """return the rule's error message"""
        return _("Marriage after death")


class EarlyMarriage(FamilyRule):
    """test if each family's parent was 'too young' at the marriage"""

    ID = 20
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, yng_mar, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.yng_mar = yng_mar
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (
            self.yng_mar,
            self.est,
        )

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        marr_date = self.obj.get_marriage_date()
        marr_date_ok = marr_date > 0

        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_birth_date = mother.get_birth_date(self.est)
        father_birth_date = father.get_birth_date(self.est)
        mother_birth_date_ok = mother_birth_date > 0
        father_birth_date_ok = father_birth_date > 0

        father_broken = (
            father_birth_date_ok
            and marr_date_ok
            and father_birth_date < marr_date
            and ((marr_date - father_birth_date) / 365 < self.yng_mar)
        )
        mother_broken = (
            mother_birth_date_ok
            and marr_date_ok
            and mother_birth_date < marr_date
            and ((marr_date - mother_birth_date) / 365 < self.yng_mar)
        )

        return father_broken or mother_broken

    def get_message(self):
        """return the rule's error message"""
        return _("Early marriage")


class LateMarriage(FamilyRule):
    """test if each family's parent was 'too old' at the marriage"""

    ID = 21
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, old_mar, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.old_mar = old_mar
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.old_mar, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        marr_date = self.obj.get_marriage_date()
        marr_date_ok = marr_date > 0

        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_birth_date = mother.get_birth_date(self.est)
        father_birth_date = father.get_birth_date(self.est)
        mother_birth_date_ok = mother_birth_date > 0
        father_birth_date_ok = father_birth_date > 0

        father_broken = (
            father_birth_date_ok
            and marr_date_ok
            and ((marr_date - father_birth_date) / 365 > self.old_mar)
        )
        mother_broken = (
            mother_birth_date_ok
            and marr_date_ok
            and ((marr_date - mother_birth_date) / 365 > self.old_mar)
        )

        return father_broken or mother_broken

    def get_message(self):
        """return the rule's error message"""
        return _("Late marriage")


class OldParent(FamilyRule):
    """test if each family's parent was 'too old' at a child's birth"""

    ID = 22
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, old_mom, old_dad, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.old_mom = old_mom
        self.old_dad = old_dad
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.old_mom, self.old_dad, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_birth_date = mother.get_birth_date(self.est)
        father_birth_date = father.get_birth_date(self.est)
        mother_birth_date_ok = mother_birth_date > 0
        father_birth_date_ok = father_birth_date > 0

        for child_ref in self.obj.get_child_ref_list():
            child = find_person(self.db, child_ref.ref)
            child_birth_date = child.get_birth_date(self.est)
            child_birth_date_ok = child_birth_date > 0
            if not child_birth_date_ok:
                continue
            father_broken = father_birth_date_ok and (
                (child_birth_date - father_birth_date) / 365 > self.old_dad
            )
            if father_broken:
                self.get_message = self.father_message
                return True

            mother_broken = mother_birth_date_ok and (
                (child_birth_date - mother_birth_date) / 365 > self.old_mom
            )
            if mother_broken:
                self.get_message = self.mother_message
                return True
        return False

    def father_message(self):
        """return the rule's error message"""
        return _("Old father")

    def mother_message(self):
        """return the rule's error message"""
        return _("Old mother")


class YoungParent(FamilyRule):
    """test if each family's parent was 'too young' at a child's birth"""

    ID = 23
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, yng_mom, yng_dad, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.yng_dad = yng_dad
        self.yng_mom = yng_mom
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.yng_mom, self.yng_dad, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_birth_date = mother.get_birth_date(self.est)
        father_birth_date = father.get_birth_date(self.est)
        mother_birth_date_ok = mother_birth_date > 0
        father_birth_date_ok = father_birth_date > 0

        for child_ref in self.obj.get_child_ref_list():
            child = find_person(self.db, child_ref.ref)
            child_birth_date = child.get_birth_date(self.est)
            child_birth_date_ok = child_birth_date > 0
            if not child_birth_date_ok:
                continue
            father_broken = father_birth_date_ok and (
                (child_birth_date - father_birth_date) / 365 < self.yng_dad
            )
            if father_broken:
                self.get_message = self.father_message
                return True

            mother_broken = mother_birth_date_ok and (
                (child_birth_date - mother_birth_date) / 365 < self.yng_mom
            )
            if mother_broken:
                self.get_message = self.mother_message
                return True
        return False

    def father_message(self):
        """return the rule's error message"""
        return _("Young father")

    def mother_message(self):
        """return the rule's error message"""
        return _("Young mother")


class UnbornParent(FamilyRule):
    """test if each family's parent was not yet born at a child's birth"""

    ID = 24
    SEVERITY = Rule.ERROR

    def __init__(self, db, obj, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.est,)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_birth_date = mother.get_birth_date(self.est)
        father_birth_date = father.get_birth_date(self.est)
        mother_birth_date_ok = mother_birth_date > 0
        father_birth_date_ok = father_birth_date > 0

        for child_ref in self.obj.get_child_ref_list():
            child = find_person(self.db, child_ref.ref)
            child_birth_date = child.get_birth_date(self.est)
            child_birth_date_ok = child_birth_date > 0
            if not child_birth_date_ok:
                continue
            father_broken = father_birth_date_ok and (
                father_birth_date > child_birth_date
            )
            if father_broken:
                self.get_message = self.father_message
                return True

            mother_broken = mother_birth_date_ok and (
                mother_birth_date > child_birth_date
            )
            if mother_broken:
                self.get_message = self.mother_message
                return True
        return False

    def father_message(self):
        """return the rule's error message"""
        return _("Unborn father")

    def mother_message(self):
        """return the rule's error message"""
        return _("Unborn mother")


class DeadParent(FamilyRule):
    """test if each family's parent was dead at a child's birth"""

    ID = 25
    SEVERITY = Rule.ERROR

    def __init__(self, db, obj, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.est,)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        mother = get_mother(self.db, self.obj)
        father = get_father(self.db, self.obj)
        mother_death_date = mother.get_death_date(self.est)
        father_death_date = father.get_death_date(self.est)
        mother_death_date_ok = mother_death_date > 0
        father_death_date_ok = father_death_date > 0

        for child_ref in self.obj.get_child_ref_list():
            child = find_person(self.db, child_ref.ref)
            child_birth_date = child.get_birth_date(self.est)
            child_birth_date_ok = child_birth_date > 0
            if not child_birth_date_ok:
                continue

            has_birth_rel_to_mother = child_ref.mrel == ChildRefType.BIRTH
            has_birth_rel_to_father = child_ref.frel == ChildRefType.BIRTH

            father_broken = (
                has_birth_rel_to_father
                and father_death_date_ok
                and ((father_death_date + 294) < child_birth_date)
            )
            if father_broken:
                self.get_message = self.father_message
                return True

            mother_broken = (
                has_birth_rel_to_mother
                and mother_death_date_ok
                and (mother_death_date < child_birth_date)
            )
            if mother_broken:
                self.get_message = self.mother_message
                return True
        return False

    def father_message(self):
        """return the rule's error message"""
        return _("Dead father")

    def mother_message(self):
        """return the rule's error message"""
        return _("Dead mother")


class LargeChildrenSpan(FamilyRule):
    """test if a family's first and last children were born far apart"""

    ID = 26
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, cb_span, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.cbs = cb_span
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.cbs, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        child_birh_dates = get_child_birth_dates(self.db, self.obj, self.est)
        child_birh_dates.sort()

        return child_birh_dates and (
            (child_birh_dates[-1] - child_birh_dates[0]) / 365 > self.cbs
        )

    def get_message(self):
        """return the rule's error message"""
        return _("Large year span for all children")


class LargeChildrenAgeDiff(FamilyRule):
    """test if any of a family's children were born far apart"""

    ID = 27
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, c_space, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.c_space = c_space
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.c_space, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        child_birh_dates = get_child_birth_dates(self.db, self.obj, self.est)
        child_birh_dates_diff = [
            child_birh_dates[i + 1] - child_birh_dates[i]
            for i in range(len(child_birh_dates) - 1)
        ]

        return child_birh_dates_diff and max(child_birh_dates_diff) / 365 > self.c_space

    def get_message(self):
        """return the rule's error message"""
        return _("Large age differences between children")


class Disconnected(PersonRule):
    """test if a person has no children and no parents"""

    ID = 28
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        return (
            len(self.obj.get_parent_family_handle_list())
            + len(self.obj.get_family_handle_list())
            == 0
        )

    def get_message(self):
        """return the rule's error message"""
        return _("Disconnected individual")


class InvalidBirthDate(PersonRule):
    """test if a person has an 'invalid' birth date"""

    ID = 29
    SEVERITY = Rule.ERROR

    def __init__(self, db, person, invdate):
        """initialize the rule"""
        PersonRule.__init__(self, db, person)
        self._invdate = invdate

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        if not self._invdate:  # should we check?
            return False
        # if so, let's get the birth date
        return self.obj.is_birth_date_invalid()

    def get_message(self):
        """return the rule's error message"""
        return _("Invalid birth date")


class InvalidDeathDate(PersonRule):
    """test if a person has an 'invalid' death date"""

    ID = 30
    SEVERITY = Rule.ERROR

    def __init__(self, db, person, invdate):
        """initialize the rule"""
        PersonRule.__init__(self, db, person)
        self._invdate = invdate

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        if not self._invdate:  # should we check?
            return False
        # if so, let's get the death date
        return self.obj.is_death_date_invalid()

    def get_message(self):
        """return the rule's error message"""
        return _("Invalid death date")


class MarriedRelation(FamilyRule):
    """test if a family has a marriage date but is not marked 'married'"""

    ID = 31
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        marr_date = self.obj.get_marriage_date()
        marr_date_ok = marr_date > 0
        married = self.obj.get_relationship() == FamilyRelType.MARRIED
        if not married and marr_date_ok:
            return self.get_message
        return False

    def get_message(self):
        """return the rule's error message"""
        return _("Marriage date but not married")


class OldAgeButNoDeath(PersonRule):
    """test if a person is 'too old' but is not shown as dead"""

    ID = 32
    SEVERITY = Rule.WARNING

    def __init__(self, db, person, old_age, est):
        """initialize the rule"""
        PersonRule.__init__(self, db, person)
        self.old_age = old_age
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.old_age, self.est)

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        birth_date = self.obj.get_birth_date(self.est)
        dead = self.obj.get_death()
        death_date = self.obj.get_death_date(True)  # or burial date
        if dead or death_date or not birth_date:
            return 0
        age = (_today - birth_date) / 365
        return age > self.old_age

    def get_message(self):
        """return the rule's error message"""
        return _("Old age but no death")


class BirthEqualsDeath(PersonRule):
    """test if a person's birth date is the same as their death date"""

    ID = 33
    SEVERITY = Rule.WARNING

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        birth_date = self.obj.get_birth_date()
        death_date = self.obj.get_death_date()
        birth_ok = birth_date > 0 if birth_date is not None else False
        death_ok = death_date > 0 if death_date is not None else False
        return death_ok and birth_ok and birth_date == death_date

    def get_message(self):
        """return the rule's error message"""
        return _("Birth date equals death date")


class BirthEqualsMarriage(PersonRule):
    """test if a person's birth date is the same as their marriage date"""

    ID = 34
    SEVERITY = Rule.ERROR

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        birth_date = self.obj.get_birth_date()
        birth_ok = birth_date > 0 if birth_date is not None else False
        for fhandle in self.obj.get_family_handle_list():
            family = find_family(self.db, fhandle)
            marr_date = family.get_marriage_date()
            marr_ok = marr_date > 0 if marr_date is not None else False
            return marr_ok and birth_ok and birth_date == marr_date

    def get_message(self):
        """return the rule's error message"""
        return _("Birth date equals marriage date")


class DeathEqualsMarriage(PersonRule):
    """test if a person's death date is the same as their marriage date"""

    ID = 35
    SEVERITY = Rule.WARNING  # it's possible

    def broken(self):
        """return boolean indicating whether this rule is violated"""
        death_date = self.obj.get_death_date()
        death_ok = death_date > 0 if death_date is not None else False
        for fhandle in self.obj.get_family_handle_list():
            family = find_family(self.db, fhandle)
            marr_date = family.get_marriage_date()
            marr_ok = marr_date > 0 if marr_date is not None else False
            return marr_ok and death_ok and death_date == marr_date

    def get_message(self):
        """return the rule's error message"""
        return _("Death date equals marriage date")


class BaptTooLate(PersonRule):
    """test if a person's baptism date is too late considering family tradition"""

    ID = 36
    SEVERITY = Rule.WARNING

    def broken(self):
        parents = self.obj.get_parent_family_handle_list()
        if len(parents) != 1:
            # only check if the person has exactly one parent family
            return False

        family = find_family(self.db, parents[0])
        if not family:
            # family not found?
            return False

        children = family.get_child_ref_list()
        if len(children) <= 1:
            # only one child? nothing to compare with...
            return False

        birth_date = self.obj.get_birth_date()
        bapt_date = self.obj.get_bapt_date()
        birth_ok = birth_date > 0 if birth_date is not None else False
        bapt_ok = bapt_date > 0 if bapt_date is not None else False
        if not birth_ok or not bapt_ok or bapt_date < birth_date:
            # return on invalid or incomplete data of the test subject
            return False
        birth_bapt_distance = bapt_date - birth_date

        child_birth_bapt_distances = []
        for childref in children:
            if int(childref.get_mother_relation()) == ChildRefType.BIRTH:
                child = find_person(self.db, childref.ref)
                if self.obj.get_gramps_id() == child.get_gramps_id():
                    continue
                birth_date = child.get_birth_date()
                bapt_date = child.get_bapt_date()
                birth_ok = birth_date > 0 if birth_date is not None else False
                bapt_ok = bapt_date > 0 if bapt_date is not None else False
                if birth_ok and bapt_ok and bapt_date >= birth_date:
                    # only collect valid and complete data
                    child_birth_bapt_distances.append(bapt_date - birth_date)

        if len(child_birth_bapt_distances) == 0:
            # only continue if we have collected some distances
            return False

        median_birth_bapt_distance = statistics.median(child_birth_bapt_distances)

        if birth_bapt_distance > median_birth_bapt_distance + 120:
            return True

        return False

    def get_message(self):
        """return the rule's error message"""
        return _("Baptism too late according to family tradition")


class BuryTooLate(PersonRule):
    """test if a person's burial date is too late"""

    ID = 37
    SEVERITY = Rule.WARNING

    def broken(self):
        death_date = self.obj.get_death_date()
        bury_date = self.obj.get_bury_date()
        death_ok = death_date > 0 if death_date is not None else False
        bury_ok = bury_date > 0 if bury_date is not None else False
        if not death_ok or not bury_ok or bury_date < death_date:
            return False

        death_bury_distance = bury_date - death_date
        if death_bury_distance > 14:
            return True

        return False

    def get_message(self):
        """return the rule's error message"""
        return _("Burial too late")


class ChildrenOrderIncorrect(FamilyRule):
    """test if children are ordered incorrectly within a family"""

    ID = 38
    SEVERITY = Rule.ERROR

    def __init__(self, db, obj, est):
        """initialize the rule"""
        FamilyRule.__init__(self, db, obj)
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.est,)

    def broken(self):
        children = self.obj.get_child_ref_list()
        if len(children) <= 1:
            # only one child? nothing to do...
            return False

        prev_birth_date = 0
        for childref in children:
            if int(childref.get_mother_relation()) == ChildRefType.BIRTH:
                child = find_person(self.db, childref.ref)
                birth_date = child.get_birth_date(self.est)
                birth_ok = birth_date > 0 if birth_date is not None else False
                if birth_ok and birth_date < prev_birth_date:
                    return True
                prev_birth_date = birth_date

        return False

    def get_message(self):
        return _("Children are not in chronological order")


class FamilyOrderIncorrect(PersonRule):
    """test if Families of a person ordered incorrectly"""

    ID = 39
    SEVERITY = Rule.WARNING

    def __init__(self, db, obj, est):
        """initialize the rule"""
        PersonRule.__init__(self, db, obj)
        self.est = est

    def _get_params(self):
        """return the rule's parameters"""
        return (self.est,)

    def broken(self):
        families = self.obj.get_family_handle_list()
        if len(families) < 2:
            # only check if the person has more than one families
            return False

        prev_compare_date = 0
        for fhandle in families:
            family = find_family(self.db, fhandle)
            if not family:
                # family not found?
                continue

            compare_date = 0
            # first try with marriage date for comparison
            marr_date = family.get_marriage_date()
            marr_ok = marr_date > 0 if marr_date is not None else False
            if marr_ok:
                compare_date = marr_date
            else:
                # if there is no, take the divorce date
                div_date = family.get_divorce_date()
                div_ok = div_date > 0 if div_date is not None else False
                if div_ok:
                    compare_date = div_date
                else:
                    # if there is no, check for the birth date of the oldest child
                    for childref in family.get_child_ref_list():
                        if int(childref.get_mother_relation()) == ChildRefType.BIRTH:
                            child = find_person(self.db, childref.ref)
                            birth_date = child.get_birth_date(self.est)
                            birth_ok = (
                                birth_date > 0 if birth_date is not None else False
                            )
                            if (
                                birth_ok
                                and birth_date < compare_date
                                or compare_date == 0
                            ):
                                compare_date = birth_date
            if compare_date != 0 and compare_date < prev_compare_date:
                return True
            prev_compare_date = compare_date
        return False

    def get_message(self):
        return _("Families are not in chronological order")


class FamilyHasEventsOfTypeUnknown(FamilyRule):
    """test if the family has events with role Unknown"""

    ID = 40
    SEVERITY = Rule.ERROR

    def broken(self):
        return self.obj.is_events_of_type_unknown()

    def get_message(self):
        return _("Family has events with role Unknown")


class PersonHasEventsOfTypeUnknown(PersonRule):
    """test if the Person has events with role Unknown"""

    ID = 41
    SEVERITY = Rule.ERROR

    def broken(self):
        return self.obj.is_events_of_type_unknown()

    def get_message(self):
        return _("Person has events with role Unknown")


class FamilyHasEventsInWrongOrder(FamilyRule):
    """test if the family has events in wrong order"""

    ID = 42
    SEVERITY = Rule.ERROR

    def broken(self):
        return self.obj.is_events_in_wrong_order()

    def get_message(self):
        return _("Family events are not in chronological order")


class PersonHasEventsInWrongOrder(PersonRule):
    """test if the person has events in wrong order"""

    ID = 43
    SEVERITY = Rule.ERROR

    def broken(self):
        return self.obj.is_events_in_wrong_order()

    def get_message(self):
        return _("Person events are not in chronological order")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Name types.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .grampstype import GrampsType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


class NameOriginType(GrampsType):
    """
    Name Origin Types

    .. attribute UNKNOWN:    Unknown origin
    .. attribute CUSTOM:     Custom user defined origin
    .. attribute NONE:       no given origin
    .. attribute INHERITED:  name was inherited from parents
    .. attribute PATRILINEAL:name was inherited from father's family name
    .. attribute MATRILINEAL:name was inherited from mother's family name
    .. attribute GIVEN:      name was bestowed on the individual
    .. attribute TAKEN:      name was chosen by the individual
    .. attribute PATRONYMIC: name is derived from father's given name
    .. attribute MATRONYMIC: name is derived from mother's given name
    .. attribute FEUDAL:     name refers to the holding of land in a fief
    .. attribute PSEUDONYM:  name is fictitious
    .. attribute OCCUPATION: name follows from the occupation of the person
    .. attribute LOCATION:   name follows from the location of the person
    """

    UNKNOWN = -1
    CUSTOM = 0
    NONE = 1
    INHERITED = 2
    GIVEN = 3
    TAKEN = 4
    PATRONYMIC = 5
    MATRONYMIC = 6
    FEUDAL = 7
    PSEUDONYM = 8
    PATRILINEAL = 9
    MATRILINEAL = 10
    OCCUPATION = 11
    LOCATION = 12

    _CUSTOM = CUSTOM
    _DEFAULT = NONE

    _DATAMAP = [
        (UNKNOWN, _("Unknown"), "Unknown "),
        (CUSTOM, _("Custom"), "Custom"),
        (NONE, "", ""),
        (INHERITED, _("Inherited", "Surname"), "Inherited"),
        (GIVEN, _("Given", "Surname"), "Given"),
        (TAKEN, _("Taken", "Surname"), "Taken"),
        (PATRONYMIC, _("Patronymic"), "Patronymic"),
        (MATRONYMIC, _("Matronymic"), "Matronymic"),
        (FEUDAL, _("Feudal", "Surname"), "Feudal"),
        (PSEUDONYM, _("Pseudonym"), "Pseudonym"),
        (PATRILINEAL, _("Patrilineal"), "Patrilineal"),
        (MATRILINEAL, _("Matrilineal"), "Matrilineal"),
        (OCCUPATION, _("Occupation"), "Occupation"),
        (LOCATION, _("Location"), "Location"),
    ]

    def __init__(self, value=None):
        GrampsType.__init__(self, value)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2020       Paul Culley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Library common to SourceView and CitationTreeView
"""
from gramps.gen.errors import HandleError


# -------------------------------------------------------------------------
#
# SourceLibView
#
# -------------------------------------------------------------------------
class LibSourceView:
    """
    This contains the common delete related methods for the views.
    It was written specifically for CitationTreeView, but works for SourceView
    as well; there just will never be an citation handles in selection.

    This must be placed before Listview in the MRO to properly override the
    included methods.  For example:
        class SourceView(LibSourceView, ListView)
    """

    def remove(self, *obj):
        """
        Method called when deleting source(s) or citations from the views.
        """
        ht_list = []
        handles = self.selected_handles()
        for hndl in handles:
            if self.dbstate.db.has_source_handle(hndl):
                ht_list.append(("Source", hndl))
            else:
                ht_list.append(("Citation", hndl))
        self.remove_selected_objects(ht_list)

    def remove_object_from_handle(
        self, obj_type, handle, trans, in_use_prompt=False, parent=None
    ):
        """
        deletes a single object from database
        """
        try:  # need this in case user selects both source and its Citations
            obj = self.dbstate.db.method("get_%s_from_handle", obj_type)(handle)
        except HandleError:
            return
        bl_list = list(self.dbstate.db.find_backlink_handles(handle))
        if in_use_prompt:
            res = self._in_use_prompt(obj, bl_list, parent=parent)
            if res != 1:  # Cancel or No
                return res
        # perfom the cleanup
        if obj_type == "Source":
            # we need to delete all back linked citations, so sort these out
            cit_list = []
            nbl_list = []
            for item in bl_list:
                if item[0] == "Citation":
                    cit_list.append(item[1])
                else:  # save any other back links for later.
                    nbl_list.append(item)
            # now lets go through citations and clean up their back-refs
            hndl_cnt = len(cit_list) / 100
            for indx, cit_hndl in enumerate(cit_list):
                # the following introduces another pass with the progressbar
                # to keep the user from wondering what is happening if there
                # are a lot of citations on the source
                self.uistate.pulse_progressbar(indx / hndl_cnt)
                cit_bl_list = list(self.dbstate.db.find_backlink_handles(cit_hndl))
                for ref_type, ref_hndl in cit_bl_list:
                    ref_obj = self.dbstate.db.method("get_%s_from_handle", ref_type)(
                        ref_hndl
                    )
                    ref_obj.remove_handle_references(obj_type, [cit_hndl])
                    self.dbstate.db.method("commit_%s", ref_type)(ref_obj, trans)
                # and delete the citation
                self.dbstate.db.remove_citation(cit_hndl, trans)
            bl_list = nbl_list  # to clean up any other back refs to source
        for ref_type, ref_hndl in bl_list:
            ref_obj = self.dbstate.db.method("get_%s_from_handle", ref_type)(ref_hndl)
            ref_obj.remove_handle_references(obj_type, [handle])
            self.dbstate.db.method("commit_%s", ref_type)(ref_obj, trans)
        # and delete the source
        self.dbstate.db.remove_source(handle, trans)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2008       Robert Cheramy <robert@cheramy.net>
# Copyright (C) 2009       Douglas S. Blank
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-2011  Nick Hall
# Copyright (C) 2013  Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Contains the interface to allow a database to get written using
Gramps' XML file format.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import time
import shutil
import os
import codecs
from xml.sax.saxutils import escape

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".WriteXML")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import URL_HOMEPAGE
from gramps.gen.lib import Date, Person
from gramps.gen.updatecallback import UpdateCallback
from gramps.gen.db.exceptions import DbWriteFailure
from gramps.version import VERSION
from gramps.gen.constfunc import win
from gramps.gui.plug.export import WriterOptionBox, WriterOptionBoxWithCompression
import gramps.plugins.lib.libgrampsxml as libgrampsxml

# -------------------------------------------------------------------------
#
# Attempt to load the GZIP library. Some version of python do not seem
# to be compiled with this available.
#
# -------------------------------------------------------------------------
try:
    import gzip

    _gzip_ok = 1
except:
    _gzip_ok = 0

# table for skipping control chars from XML except 09, 0A, 0D
strip_dict = dict.fromkeys(list(range(9)) + list(range(11, 13)) + list(range(14, 32)))


def escxml(d):
    return (
        escape(
            d,
            {
                '"': "&quot;",
                "<": "&lt;",
                ">": "&gt;",
            },
        )
        if d
        else ""
    )


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class GrampsXmlWriter(UpdateCallback):
    """
    Writes a database to the XML file.
    """

    def __init__(self, db, strip_photos=0, compress=1, version="unknown", user=None):
        """
        Initialize, but does not write, an XML file.

        db - database to write
        strip_photos - remove paths off of media object paths
        >              0: do not touch the paths
        >              1: remove everything expect the filename (eg gpkg)
        >              2: remove leading slash (quick write)
        compress - attempt to compress the database
        """
        UpdateCallback.__init__(self, user.callback)
        self.user = user
        self.compress = compress
        if not _gzip_ok:
            self.compress = False
        self.db = db
        self.strip_photos = strip_photos
        self.version = version

        self.status = None

    def write(self, filename):
        """
        Write the database to the specified file.
        """
        if filename == "-":
            import sys

            try:
                g = sys.stdout.buffer
            except:
                g = sys.stdout
            self.compress = False
        else:
            base = os.path.dirname(filename)
            if os.path.isdir(base):
                if not os.access(base, os.W_OK) or not os.access(base, os.R_OK):
                    raise DbWriteFailure(
                        _("Failure writing %s") % filename,
                        _(
                            "The database cannot be saved because you do "
                            "not have permission to write to the directory. "
                            "Please make sure you have write access to the "
                            "directory and try again."
                        ),
                    )
                    return 0
            else:
                raise DbWriteFailure(
                    _("No directory"),
                    _(
                        "There is no directory %s.\n\n"
                        "Please select another directory "
                        "or create it."
                    )
                    % base,
                )
                return 0

            if os.path.exists(filename):
                if not os.access(filename, os.W_OK):
                    raise DbWriteFailure(
                        _("Failure writing %s") % filename,
                        _(
                            "The database cannot be saved because you do "
                            "not have permission to write to the file. "
                            "Please make sure you have write access to the "
                            "file and try again."
                        ),
                    )
                    return 0

            self.fileroot = os.path.dirname(filename)
            try:
                if self.compress and _gzip_ok:
                    try:
                        g = gzip.open(filename, "wb")
                    except:
                        g = open(filename, "wb")
                else:
                    g = open(filename, "wb")
            except IOError as msg:
                LOG.warning(str(msg))
                raise DbWriteFailure(_("Failure writing %s") % filename, str(msg))
                return 0

        self.g = codecs.getwriter("utf8")(g)

        self.write_xml_data()
        if filename != "-":
            g.close()
        return 1

    def write_handle(self, handle):
        """
        Write the database to the specified file handle.
        """

        if self.compress and _gzip_ok:
            try:
                g = gzip.GzipFile(mode="wb", fileobj=handle)
            except:
                g = handle
        else:
            g = handle

        self.g = codecs.getwriter("utf8")(g)

        self.write_xml_data()
        g.close()
        return 1

    def write_xml_data(self):
        date = time.localtime(time.time())
        owner = self.db.get_researcher()

        person_len = self.db.get_number_of_people()
        family_len = self.db.get_number_of_families()
        event_len = self.db.get_number_of_events()
        citation_len = self.db.get_number_of_citations()
        source_len = self.db.get_number_of_sources()
        place_len = self.db.get_number_of_places()
        repo_len = self.db.get_number_of_repositories()
        obj_len = self.db.get_number_of_media()
        note_len = self.db.get_number_of_notes()
        tag_len = self.db.get_number_of_tags()

        total_steps = (
            person_len
            + family_len
            + event_len
            + citation_len
            + source_len
            + place_len
            + repo_len
            + obj_len
            + note_len
            + tag_len
        )

        self.set_total(total_steps)

        self.g.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        self.g.write(
            "<!DOCTYPE database "
            'PUBLIC "-//Gramps//DTD Gramps XML %s//EN"\n'
            '"%sxml/%s/grampsxml.dtd">\n'
            % (
                libgrampsxml.GRAMPS_XML_VERSION,
                URL_HOMEPAGE,
                libgrampsxml.GRAMPS_XML_VERSION,
            )
        )
        self.g.write(
            '<database xmlns="%sxml/%s/">\n'
            % (URL_HOMEPAGE, libgrampsxml.GRAMPS_XML_VERSION)
        )
        self.g.write("  <header>\n")
        self.g.write('    <created date="%04d-%02d-%02d"' % date[:3])
        self.g.write(' version="' + self.version + '"')
        self.g.write("/>\n")
        self.g.write("    <researcher>\n")
        self.write_line("resname", owner.get_name(), 3)
        self.write_line("resaddr", owner.get_address(), 3)
        self.write_line("reslocality", owner.get_locality(), 3)
        self.write_line("rescity", owner.get_city(), 3)
        self.write_line("resstate", owner.get_state(), 3)
        self.write_line("rescountry", owner.get_country(), 3)
        self.write_line("respostal", owner.get_postal_code(), 3)
        self.write_line("resphone", owner.get_phone(), 3)
        self.write_line("resemail", owner.get_email(), 3)
        self.g.write("    </researcher>\n")
        self.write_metadata()
        self.g.write("  </header>\n")

        # First write name formats: we need to know all formats
        # by the time we get to person's names
        self.write_name_formats()

        # Write table objects
        if tag_len > 0:
            self.g.write("  <tags>\n")
            for key in sorted(self.db.get_tag_handles()):
                tag = self.db.get_tag_from_handle(key)
                if tag:
                    self.write_tag(tag, 2)
                self.update()
            self.g.write("  </tags>\n")

        # Write primary objects
        if event_len > 0:
            self.g.write("  <events>\n")
            for handle in sorted(self.db.get_event_handles()):
                event = self.db.get_event_from_handle(handle)
                if event:
                    self.write_event(event, 2)
                self.update()
            self.g.write("  </events>\n")

        if person_len > 0:
            self.g.write("  <people")
            person = self.db.get_default_person()
            if person:
                self.g.write(' home="_%s"' % person.handle)
            self.g.write(">\n")

            for handle in sorted(self.db.get_person_handles()):
                person = self.db.get_person_from_handle(handle)
                if person:
                    self.write_person(person, 2)
                self.update()
            self.g.write("  </people>\n")

        if family_len > 0:
            self.g.write("  <families>\n")
            for handle in sorted(self.db.iter_family_handles()):
                family = self.db.get_family_from_handle(handle)
                if family:
                    self.write_family(family, 2)
                self.update()
            self.g.write("  </families>\n")

        if citation_len > 0:
            self.g.write("  <citations>\n")
            for handle in sorted(self.db.get_citation_handles()):
                citation = self.db.get_citation_from_handle(handle)
                if citation:
                    self.write_citation(citation, 2)
                self.update()
            self.g.write("  </citations>\n")

        if source_len > 0:
            self.g.write("  <sources>\n")
            for handle in sorted(self.db.get_source_handles()):
                source = self.db.get_source_from_handle(handle)
                if source:
                    self.write_source(source, 2)
                self.update()
            self.g.write("  </sources>\n")

        if place_len > 0:
            self.g.write("  <places>\n")
            for key in sorted(self.db.get_place_handles()):
                # try:
                place = self.db.get_place_from_handle(key)
                if place:
                    self.write_place_obj(place, 2)
                self.update()
            self.g.write("  </places>\n")

        if obj_len > 0:
            self.g.write("  <objects>\n")
            for handle in sorted(self.db.get_media_handles()):
                obj = self.db.get_media_from_handle(handle)
                if obj:
                    self.write_object(obj, 2)
                self.update()
            self.g.write("  </objects>\n")

        if repo_len > 0:
            self.g.write("  <repositories>\n")
            for key in sorted(self.db.get_repository_handles()):
                repo = self.db.get_repository_from_handle(key)
                if repo:
                    self.write_repository(repo, 2)
                self.update()
            self.g.write("  </repositories>\n")

        if note_len > 0:
            self.g.write("  <notes>\n")
            for key in sorted(self.db.get_note_handles()):
                note = self.db.get_note_from_handle(key)
                if note:
                    self.write_note(note, 2)
                self.update()
            self.g.write("  </notes>\n")

        # Data is written, now write bookmarks.
        self.write_bookmarks()
        self.write_namemaps()

        self.g.write("</database>\n")

    #        self.status.end()
    #        self.status = None

    def write_metadata(self):
        """Method to write out metadata of the database"""
        mediapath = self.db.get_mediapath()
        if mediapath is not None:
            self.write_line("mediapath", mediapath, 2)

    def write_namemaps(self):
        group_map = self.db.get_name_group_keys()
        name_len = len(group_map)

        if name_len > 0:
            self.g.write("  <namemaps>\n")
            for key in group_map:
                value = self.db.get_name_group_mapping(key)
                self.g.write(
                    '    <map type="group_as" key="%s" value="%s"/>\n'
                    % (self.fix(key), self.fix(value))
                )
            self.g.write("  </namemaps>\n")

    def write_bookmarks(self):
        bm_person_len = len(self.db.bookmarks.get())
        bm_family_len = len(self.db.family_bookmarks.get())
        bm_event_len = len(self.db.event_bookmarks.get())
        bm_source_len = len(self.db.source_bookmarks.get())
        bm_citation_len = len(self.db.citation_bookmarks.get())
        bm_place_len = len(self.db.place_bookmarks.get())
        bm_repo_len = len(self.db.repo_bookmarks.get())
        bm_obj_len = len(self.db.media_bookmarks.get())
        bm_note_len = len(self.db.note_bookmarks.get())

        bm_len = (
            bm_person_len
            + bm_family_len
            + bm_event_len
            + bm_source_len
            + bm_place_len
            + bm_repo_len
            + bm_citation_len
            + bm_obj_len
            + bm_note_len
        )

        if bm_len > 0:
            self.g.write("  <bookmarks>\n")

            for handle in self.db.get_bookmarks().get():
                self.g.write('    <bookmark target="person" hlink="_%s"/>\n' % handle)
            for handle in self.db.get_family_bookmarks().get():
                self.g.write('    <bookmark target="family" hlink="_%s"/>\n' % handle)
            for handle in self.db.get_event_bookmarks().get():
                self.g.write('    <bookmark target="event" hlink="_%s"/>\n' % handle)
            for handle in self.db.get_source_bookmarks().get():
                self.g.write('    <bookmark target="source" hlink="_%s"/>\n' % handle)
            for handle in self.db.get_citation_bookmarks().get():
                self.g.write('    <bookmark target="citation" hlink="_%s"/>\n' % handle)
            for handle in self.db.get_place_bookmarks().get():
                self.g.write('    <bookmark target="place" hlink="_%s"/>\n' % handle)
            for handle in self.db.get_media_bookmarks().get():
                self.g.write('    <bookmark target="media" hlink="_%s"/>\n' % handle)
            for handle in self.db.get_repo_bookmarks().get():
                self.g.write(
                    '    <bookmark target="repository" hlink="_%s"/>\n' % handle
                )
            for handle in self.db.get_note_bookmarks().get():
                self.g.write('    <bookmark target="note" hlink="_%s"/>\n' % handle)

            self.g.write("  </bookmarks>\n")

    def write_name_formats(self):
        if len(self.db.name_formats) > 0:
            self.g.write("  <name-formats>\n")
            for number, name, fmt_str, active in self.db.name_formats:
                self.g.write(
                    '%s<format number="%d" name="%s" '
                    'fmt_str="%s" active="%d"/>\n'
                    % ("    ", number, escxml(name), escxml(fmt_str), int(active))
                )
            self.g.write("  </name-formats>\n")

    def write_tag(self, tag, index=2):
        """
        Write a tag definition.
        """
        if not tag:
            return

        self.write_table_tag("tag", tag, index, close=False)
        self.g.write(' name="%s"' % escxml(tag.get_name()))
        self.g.write(' color="%s"' % tag.get_color())
        self.g.write(' priority="%d"' % tag.get_priority())
        self.g.write("/>\n")

    def fix(self, line):
        l = str(line)
        l = l.strip().translate(strip_dict)
        return escxml(l)

    def write_note_list(self, note_list, indent=0):
        for handle in note_list:
            self.write_ref("noteref", handle, indent)

    def write_note(self, note, index=2):
        if not note:
            return

        self.write_primary_tag("note", note, index, close=False)

        ntype = escxml(note.get_type().xml_str())
        format = note.get_format()
        text = note.get_styledtext()
        styles = text.get_tags()
        text = str(text)

        self.g.write(' type="%s"' % ntype)
        if format != note.FLOWED:
            self.g.write(' format="%d"' % format)
        self.g.write(">\n")

        self.write_text("text", text, index + 1)

        if styles:
            self.write_styles(styles, index + 1)

        for tag_handle in note.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("  " * index + "</note>\n")

    def write_styles(self, styles, index=3):
        for style in styles:
            name = style.name.xml_str()
            value = style.value

            self.g.write("  " * index + '<style name="%s"' % name)
            if value:
                self.g.write(' value="%s"' % escxml(str(value)))
            self.g.write(">\n")

            for start, end in style.ranges:
                self.g.write(
                    ("  " * (index + 1))
                    + '<range start="%d" end="%d"/>\n' % (start, end)
                )

            self.g.write("  " * index + "</style>\n")

    def write_text(self, val, text, indent=0):
        if not text:
            return

        if indent:
            self.g.write("  " * indent)

        self.g.write("<%s>" % val)
        self.g.write(escxml(str(text).translate(strip_dict)))
        self.g.write("</%s>\n" % val)

    def write_person(self, person, index=1):
        sp = "  " * index
        self.write_primary_tag("person", person, index)
        if person.get_gender() == Person.MALE:
            self.write_line("gender", "M", index + 1)
        elif person.get_gender() == Person.FEMALE:
            self.write_line("gender", "F", index + 1)
        elif person.get_gender() == Person.OTHER:
            self.write_line("gender", "X", index + 1)
        else:
            self.write_line("gender", "U", index + 1)
        self.dump_name(person.get_primary_name(), False, index + 1)
        for name in person.get_alternate_names():
            self.dump_name(name, True, index + 1)

        # self.dump_event_ref(person.get_birth_ref(),index+1)
        # self.dump_event_ref(person.get_death_ref(),index+1)
        for event_ref in person.get_event_ref_list():
            self.dump_event_ref(event_ref, index + 1)

        for lds_ord in person.lds_ord_list:
            self.dump_ordinance(lds_ord, index + 1)

        self.write_media_list(person.get_media_list(), index + 1)

        self.write_address_list(person, index + 1)
        self.write_attribute_list(person.get_attribute_list())
        self.write_url_list(person.get_url_list(), index + 1)

        for family_handle in person.get_parent_family_handle_list():
            self.write_ref("childof", family_handle, index + 1)

        for family_handle in person.get_family_handle_list():
            self.write_ref("parentin", family_handle, index + 1)

        for person_ref in person.get_person_ref_list():
            self.dump_person_ref(person_ref, index + 1)

        self.write_note_list(person.get_note_list(), index + 1)

        for citation_handle in person.get_citation_list():
            self.write_ref("citationref", citation_handle, index + 1)

        for tag_handle in person.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</person>\n" % sp)

    def write_family(self, family, index=1):
        sp = "  " * index
        self.write_family_handle(family, index)
        fhandle = family.get_father_handle()
        mhandle = family.get_mother_handle()
        if fhandle:
            self.write_ref("father", fhandle, index + 1)
        if mhandle:
            self.write_ref("mother", mhandle, index + 1)
        for event_ref in family.get_event_ref_list():
            self.dump_event_ref(event_ref, 3)
        for lds_ord in family.lds_ord_list:
            self.dump_ordinance(lds_ord, index + 1)

        self.write_media_list(family.get_media_list(), index + 1)

        for child_ref in family.get_child_ref_list():
            self.dump_child_ref(child_ref, index + 1)
        self.write_attribute_list(family.get_attribute_list())
        self.write_note_list(family.get_note_list(), index + 1)
        for citation_handle in family.get_citation_list():
            self.write_ref("citationref", citation_handle, index + 1)

        for tag_handle in family.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</family>\n" % sp)

    def write_citation(self, citation, index=1):
        sp = "  " * index
        self.write_primary_tag("citation", citation, index)
        self.write_date(citation.get_date_object(), index + 1)
        self.write_line("page", citation.get_page(), index + 1)
        self.write_line_always("confidence", citation.get_confidence_level(), index + 1)
        self.write_note_list(citation.get_note_list(), index + 1)
        self.write_media_list(citation.get_media_list(), index + 1)
        self.write_srcattribute_list(citation.get_attribute_list(), index + 1)
        self.write_ref("sourceref", citation.get_reference_handle(), index + 1)

        for tag_handle in citation.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</citation>\n" % sp)

    def write_source(self, source, index=1):
        sp = "  " * index
        self.write_primary_tag("source", source, index)
        self.write_force_line("stitle", source.get_title(), index + 1)
        self.write_line("sauthor", source.get_author(), index + 1)
        self.write_line("spubinfo", source.get_publication_info(), index + 1)
        self.write_line("sabbrev", source.get_abbreviation(), index + 1)
        self.write_note_list(source.get_note_list(), index + 1)
        self.write_media_list(source.get_media_list(), index + 1)
        self.write_srcattribute_list(source.get_attribute_list(), index + 1)
        self.write_reporef_list(source.get_reporef_list(), index + 1)

        for tag_handle in source.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</source>\n" % sp)

    def write_repository(self, repo, index=1):
        sp = "  " * index
        self.write_primary_tag("repository", repo, index)
        # name
        self.write_line("rname", repo.name, index + 1)
        rtype = repo.type.xml_str()
        if rtype:
            self.write_line("type", rtype, index + 1)
        # address list
        self.write_address_list(repo, index + 1)
        # url list
        self.write_url_list(repo.get_url_list(), index + 1)
        self.write_note_list(repo.get_note_list(), index + 1)

        for tag_handle in repo.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</repository>\n" % sp)

    def write_address_list(self, obj, index=1):
        if len(obj.get_address_list()) == 0:
            return
        sp = "  " * index
        for address in obj.get_address_list():
            self.g.write("%s<address%s>\n" % (sp, conf_priv(address)))
            self.write_date(address.get_date_object(), index + 1)
            self.write_line("street", address.get_street(), index + 1)
            self.write_line("locality", address.get_locality(), index + 1)
            self.write_line("city", address.get_city(), index + 1)
            self.write_line("county", address.get_county(), index + 1)
            self.write_line("state", address.get_state(), index + 1)
            self.write_line("country", address.get_country(), index + 1)
            self.write_line("postal", address.get_postal_code(), index + 1)
            self.write_line("phone", address.get_phone(), index + 1)
            self.write_note_list(address.get_note_list(), index + 1)
            for citation_handle in address.get_citation_list():
                self.write_ref("citationref", citation_handle, index + 1)
            self.g.write("%s</address>\n" % sp)

    def dump_person_ref(self, personref, index=1):
        if not personref or not personref.ref:
            return
        sp = "  " * index
        priv_text = conf_priv(personref)
        rel_text = ' rel="%s"' % escxml(personref.get_relation())

        citation_list = personref.get_citation_list()
        nreflist = personref.get_note_list()
        if len(citation_list) + len(nreflist) == 0:
            self.write_ref(
                "personref",
                personref.ref,
                index,
                close=True,
                extra_text=priv_text + rel_text,
            )
        else:
            self.write_ref(
                "personref",
                personref.ref,
                index,
                close=False,
                extra_text=priv_text + rel_text,
            )
            for citation_handle in citation_list:
                self.write_ref("citationref", citation_handle, index + 1)

            self.write_note_list(nreflist, index + 1)
            self.g.write("%s</personref>\n" % sp)

    def dump_child_ref(self, childref, index=1):
        if not childref or not childref.ref:
            return
        sp = "  " * index
        priv_text = conf_priv(childref)
        if childref.frel.is_default():
            frel_text = ""
        else:
            frel_text = ' frel="%s"' % escxml(childref.frel.xml_str())
        if childref.mrel.is_default():
            mrel_text = ""
        else:
            mrel_text = ' mrel="%s"' % escxml(childref.mrel.xml_str())
        citation_list = childref.get_citation_list()
        nreflist = childref.get_note_list()
        if len(citation_list) + len(nreflist) == 0:
            self.write_ref(
                "childref",
                childref.ref,
                index,
                close=True,
                extra_text=priv_text + mrel_text + frel_text,
            )
        else:
            self.write_ref(
                "childref",
                childref.ref,
                index,
                close=False,
                extra_text=priv_text + mrel_text + frel_text,
            )
            for citation_handle in citation_list:
                self.write_ref("citationref", citation_handle, index + 1)
            self.write_note_list(nreflist, index + 1)
            self.g.write("%s</childref>\n" % sp)

    def dump_event_ref(self, eventref, index=1):
        if not eventref or not eventref.ref:
            return
        sp = "  " * index
        priv_text = conf_priv(eventref)
        role = escxml(eventref.role.xml_str())
        if role:
            role_text = ' role="%s"' % role
        else:
            role_text = ""

        attribute_list = eventref.get_attribute_list()
        citation_list = eventref.get_citation_list()
        note_list = eventref.get_note_list()
        if len(citation_list) + len(attribute_list) + len(note_list) == 0:
            self.write_ref(
                "eventref",
                eventref.ref,
                index,
                close=True,
                extra_text=priv_text + role_text,
            )
        else:
            self.write_ref(
                "eventref",
                eventref.ref,
                index,
                close=False,
                extra_text=priv_text + role_text,
            )
            self.write_attribute_list(attribute_list, index + 1)
            self.write_note_list(note_list, index + 1)
            for citation_handle in citation_list:
                self.write_ref("citationref", citation_handle, index + 1)
            self.g.write("%s</eventref>\n" % sp)

    def dump_place_ref(self, placeref, index=1):
        sp = "  " * index
        date = placeref.get_date_object()
        if date.is_empty():
            self.write_ref("placeref", placeref.ref, index, close=True)
        else:
            self.write_ref("placeref", placeref.ref, index, close=False)
            self.write_date(date, index + 1)
            self.g.write("%s</placeref>\n" % sp)

    def dump_place_name(self, place_name, index=1):
        sp = "  " * index
        value = place_name.get_value()
        date = place_name.get_date_object()
        lang = place_name.get_language()
        self.g.write('%s<pname value="%s"' % (sp, self.fix(value)))
        if lang:
            self.g.write(' lang="%s"' % self.fix(lang))
        if date.is_empty():
            self.g.write("/>\n")
        else:
            self.g.write(">\n")
            self.write_date(date, index + 1)
            self.g.write("%s</pname>\n" % sp)

    def write_event(self, event, index=1):
        if not event:
            return

        self.write_primary_tag("event", event, 2)

        sp = "  " * index
        etype = event.get_type().xml_str()
        self.g.write("  %s<type>%s</type>\n" % (sp, self.fix(etype)))
        self.write_date(event.get_date_object(), index + 1)
        self.write_ref("place", event.get_place_handle(), index + 1)
        self.write_line("description", event.get_description(), index + 1)
        self.write_attribute_list(event.get_attribute_list(), index + 1)
        self.write_note_list(event.get_note_list(), index + 1)

        for citation_handle in event.get_citation_list():
            self.write_ref("citationref", citation_handle, index + 1)
        self.write_media_list(event.get_media_list(), index + 1)

        for tag_handle in event.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</event>\n" % sp)

    def dump_ordinance(self, ord, index=1):
        name = ord.type2xml()

        sp = "  " * index
        sp2 = "  " * (index + 1)

        priv = conf_priv(ord)
        self.g.write('%s<lds_ord type="%s"%s>\n' % (sp, name, priv))
        dateobj = ord.get_date_object()
        if dateobj and not dateobj.is_empty():
            self.write_date(dateobj, index + 1)
        if ord.get_temple():
            self.g.write('%s<temple val="%s"/>\n' % (sp2, self.fix(ord.get_temple())))
        self.write_ref("place", ord.get_place_handle(), index + 1)
        if ord.get_status() != 0:
            self.g.write('%s<status val="%s"/>\n' % (sp2, ord.status2xml()))
        if ord.get_family_handle():
            self.g.write(
                '%s<sealed_to hlink="%s"/>\n' % (sp2, "_" + ord.get_family_handle())
            )
        self.write_note_list(ord.get_note_list(), index + 1)
        for citation_handle in ord.get_citation_list():
            self.write_ref("citationref", citation_handle, index + 1)
        self.g.write("%s</lds_ord>\n" % sp)

    def write_ref(self, tagname, handle, index=1, close=True, extra_text=""):
        if handle:
            if close:
                close_tag = "/"
            else:
                close_tag = ""
            sp = "  " * index
            self.g.write(
                '%s<%s hlink="_%s"%s%s>\n'
                % (sp, tagname, handle, extra_text, close_tag)
            )

    def write_primary_tag(self, tagname, obj, index=1, close=True):
        """
        Write the tag attributes common to all primary objects.
        """
        if not obj:
            return
        priv_text = conf_priv(obj)
        id_text = ' id="%s"' % escxml(obj.gramps_id)

        self.write_table_tag(tagname, obj, index, False)
        self.g.write(id_text + priv_text)
        if close:
            self.g.write(">\n")

    def write_table_tag(self, tagname, obj, index=1, close=True):
        """
        Write the tag attributes common to all table objects.
        """
        if not obj:
            return
        sp = "  " * index
        try:
            change_text = ' change="%d"' % obj.get_change_time()
        except:
            change_text = ' change="%d"' % 0

        handle_text = ' handle="_%s"' % obj.get_handle()

        obj_text = "%s<%s" % (sp, tagname)
        self.g.write(obj_text + handle_text + change_text)
        if close:
            self.g.write(">\n")

    def write_family_handle(self, family, index=1):
        sp = "  " * index
        self.write_primary_tag("family", family, index)
        if family:
            rel = escxml(family.get_relationship().xml_str())
            if rel != "":
                self.g.write('  %s<rel type="%s"/>\n' % (sp, rel))

    def write_surname(self, surname, indent=1):
        """
        Writes a surname of the name
        """
        pre = surname.get_prefix()
        con = surname.get_connector()
        nam = surname.get_surname()
        der = surname.get_origintype().xml_str()
        pri = surname.get_primary()
        self.g.write("%s<surname" % ("  " * indent))
        if pre:
            self.g.write(' prefix="%s"' % escxml(pre))
        if not pri:
            self.g.write(' prim="0"')
        if con:
            self.g.write(' connector="%s"' % escxml(con))
        if der:
            self.g.write(' derivation="%s"' % escxml(der))

        self.g.write(">%s</surname>\n" % self.fix(nam))

    def write_line(self, tagname, value, indent=1):
        if value:
            self.g.write(
                "%s<%s>%s</%s>\n" % ("  " * indent, tagname, self.fix(value), tagname)
            )

    def write_line_nofix(self, tagname, value, indent=1):
        """Writes a line, but does not escape characters.
        Use this instead of write_line if the value is already fixed,
        this avoids &amp; becoming &amp;amp;
        """
        if value:
            self.g.write("%s<%s>%s</%s>\n" % ("  " * indent, tagname, value, tagname))

    def write_line_always(self, tagname, value, indent=1):
        """Writes a line, always, even with a zero value."""
        self.g.write(
            "%s<%s>%s</%s>\n" % ("  " * indent, tagname, self.fix(value), tagname)
        )

    def get_iso_date(self, date):
        if date[2] == 0:
            y = "????"
        else:
            y = "%04d" % date[2]

        if date[1] == 0:
            if date[0] == 0:
                m = ""
            else:
                m = "-??"
        else:
            m = "-%02d" % (date[1])
        if date[0] == 0:
            d = ""
        else:
            d = "-%02d" % date[0]
        ret = "%s%s%s" % (y, m, d)
        # If the result does not contain anything beyond dashes
        # and question marks then it's as good as empty
        if ret.replace("-", "").replace("?", "") == "":
            ret = ""
        return ret

    def write_date(self, date, indent=1):
        sp = "  " * indent

        cal = date.get_calendar()
        if cal != Date.CAL_GREGORIAN:
            calstr = ' cformat="%s"' % Date.calendar_names[cal]
        else:
            calstr = ""

        qual = date.get_quality()
        if qual == Date.QUAL_ESTIMATED:
            qual_str = ' quality="estimated"'
        elif qual == Date.QUAL_CALCULATED:
            qual_str = ' quality="calculated"'
        else:
            qual_str = ""

        dualdated = date.get_slash()
        if dualdated:
            dualdated_str = ' dualdated="1"'
        else:
            dualdated_str = ""

        newyear = date.newyear_to_str()
        if newyear:
            newyear_str = ' newyear="%s"' % newyear
        else:
            newyear_str = ""

        mode = date.get_modifier()

        if date.is_compound():
            if mode == Date.MOD_RANGE:
                tagname = "daterange"
            else:
                tagname = "datespan"

            d1 = self.get_iso_date(date.get_start_date())
            d2 = self.get_iso_date(date.get_stop_date())
            if d1 != "" or d2 != "":
                self.g.write(
                    '%s<%s start="%s" stop="%s"%s%s%s%s/>\n'
                    % (
                        sp,
                        tagname,
                        d1,
                        d2,
                        qual_str,
                        calstr,
                        dualdated_str,
                        newyear_str,
                    )
                )
        elif mode != Date.MOD_TEXTONLY:
            date_str = self.get_iso_date(date.get_start_date())
            if date_str == "":
                return

            if mode == Date.MOD_BEFORE:
                mode_str = ' type="before"'
            elif mode == Date.MOD_AFTER:
                mode_str = ' type="after"'
            elif mode == Date.MOD_ABOUT:
                mode_str = ' type="about"'
            elif mode == Date.MOD_FROM:
                mode_str = ' type="from"'
            elif mode == Date.MOD_TO:
                mode_str = ' type="to"'
            else:
                mode_str = ""

            self.g.write(
                '%s<dateval val="%s"%s%s%s%s%s/>\n'
                % (sp, date_str, mode_str, qual_str, calstr, dualdated_str, newyear_str)
            )
        else:
            self.g.write('%s<datestr val="%s"/>\n' % (sp, self.fix(date.get_text())))

    def write_force_line(self, label, value, indent=1):
        if value is not None:
            self.g.write(
                "%s<%s>%s</%s>\n" % ("  " * indent, label, self.fix(value), label)
            )

    def dump_name(self, name, alternative=False, index=1):
        sp = "  " * index
        name_type = name.get_type().xml_str()
        # bug 9242
        if len(name.get_first_name().splitlines()) != 1:
            firstname = "".join(name.get_first_name().splitlines())
        else:
            firstname = name.get_first_name()
        self.g.write("%s<name" % sp)
        if alternative:
            self.g.write(' alt="1"')
        if name_type:
            self.g.write(' type="%s"' % escxml(name_type))
        if name.get_privacy() != 0:
            self.g.write(' priv="%d"' % name.get_privacy())
        if name.get_sort_as() != 0:
            self.g.write(' sort="%d"' % name.get_sort_as())
        if name.get_display_as() != 0:
            self.g.write(' display="%d"' % name.get_display_as())
        self.g.write(">\n")
        self.write_line("first", firstname, index + 1)
        self.write_line("call", name.get_call_name(), index + 1)
        for surname in name.get_surname_list():
            self.write_surname(surname, index + 1)
        self.write_line("suffix", name.get_suffix(), index + 1)
        self.write_line("title", name.get_title(), index + 1)
        self.write_line("nick", name.get_nick_name(), index + 1)
        self.write_line("familynick", name.get_family_nick_name(), index + 1)
        self.write_line("group", name.get_group_as(), index + 1)
        if name.date:
            self.write_date(name.date, 4)
        self.write_note_list(name.get_note_list(), index + 1)
        for citation_handle in name.get_citation_list():
            self.write_ref("citationref", citation_handle, index + 1)

        self.g.write("%s</name>\n" % sp)

    def append_value(self, orig, val):
        if orig:
            return "%s, %s" % (orig, val)
        else:
            return val

    def build_place_title(self, loc):
        "Builds a title from a location"
        street = self.fix(loc.get_street())
        locality = self.fix(loc.get_locality())
        city = self.fix(loc.get_city())
        parish = self.fix(loc.get_parish())
        county = self.fix(loc.get_county())
        state = self.fix(loc.get_state())
        country = self.fix(loc.get_country())

        value = ""

        if street:
            value = street
        if locality:
            value = self.append_value(value, locality)
        if city:
            value = self.append_value(value, city)
        if parish:
            value = self.append_value(value, parish)
        if county:
            value = self.append_value(value, county)
        if state:
            value = self.append_value(value, state)
        if country:
            value = self.append_value(value, country)
        return value

    def dump_location(self, loc):
        "Writes the location information to the output file"
        if loc.is_empty():
            return
        street = self.fix(loc.get_street())
        locality = self.fix(loc.get_locality())
        city = self.fix(loc.get_city())
        parish = self.fix(loc.get_parish())
        county = self.fix(loc.get_county())
        state = self.fix(loc.get_state())
        country = self.fix(loc.get_country())
        zip_code = self.fix(loc.get_postal_code())
        phone = self.fix(loc.get_phone())

        self.g.write("      <location")
        if street:
            self.g.write(' street="%s"' % street)
        if locality:
            self.g.write(' locality="%s"' % locality)
        if city:
            self.g.write(' city="%s"' % city)
        if parish:
            self.g.write(' parish="%s"' % parish)
        if county:
            self.g.write(' county="%s"' % county)
        if state:
            self.g.write(' state="%s"' % state)
        if country:
            self.g.write(' country="%s"' % country)
        if zip_code:
            self.g.write(' postal="%s"' % zip_code)
        if phone:
            self.g.write(' phone="%s"' % phone)
        self.g.write("/>\n")

    def write_attribute_list(self, list, indent=3):
        sp = "  " * indent
        for attr in list:
            self.g.write(
                '%s<attribute%s type="%s" value="%s"'
                % (
                    sp,
                    conf_priv(attr),
                    escxml(attr.get_type().xml_str()),
                    self.fix(attr.get_value()),
                )
            )
            citation_list = attr.get_citation_list()
            nlist = attr.get_note_list()
            if (len(nlist) + len(citation_list)) == 0:
                self.g.write("/>\n")
            else:
                self.g.write(">\n")
                for citation_handle in citation_list:
                    self.write_ref("citationref", citation_handle, indent + 1)
                self.write_note_list(attr.get_note_list(), indent + 1)
                self.g.write("%s</attribute>\n" % sp)

    def write_srcattribute_list(self, list, indent=3):
        sp = "  " * indent
        for srcattr in list:
            self.g.write(
                '%s<srcattribute%s type="%s" value="%s"'
                % (
                    sp,
                    conf_priv(srcattr),
                    escxml(srcattr.get_type().xml_str()),
                    self.fix(srcattr.get_value()),
                )
            )
            self.g.write("/>\n")

    def write_media_list(self, list, indent=3):
        sp = "  " * indent
        for photo in list:
            mobj_id = photo.get_reference_handle()
            self.g.write('%s<objref hlink="%s"' % (sp, "_" + mobj_id))
            if photo.get_privacy():
                self.g.write(' priv="1"')
            proplist = photo.get_attribute_list()
            citation_list = photo.get_citation_list()
            nreflist = photo.get_note_list()
            rect = photo.get_rectangle()
            if rect is not None:
                corner1_x = rect[0]
                corner1_y = rect[1]
                corner2_x = rect[2]
                corner2_y = rect[3]
                if corner1_x is None:
                    corner1_x = 0
                if corner1_y is None:
                    corner1_y = 0
                if corner2_x is None:
                    corner2_x = 100
                if corner2_y is None:
                    corner2_y = 100
                # don't output not set rectangle
                if (
                    corner1_x == corner1_y == corner2_x == corner2_y == 0
                    or corner1_x == corner1_y == 0
                    and corner2_x == corner2_y == 100
                ):
                    rect = None
            if len(proplist) + len(nreflist) + len(citation_list) == 0 and rect is None:
                self.g.write("/>\n")
            else:
                self.g.write(">\n")
                if rect is not None:
                    self.g.write(
                        ' %s<region corner1_x="%d" corner1_y="%d" '
                        'corner2_x="%d" corner2_y="%d"/>\n'
                        % (sp, corner1_x, corner1_y, corner2_x, corner2_y)
                    )
                self.write_attribute_list(proplist, indent + 1)
                for citation_handle in citation_list:
                    self.write_ref("citationref", citation_handle, indent + 1)
                self.write_note_list(nreflist, indent + 1)
                self.g.write("%s</objref>\n" % sp)

    def write_reporef_list(self, rrlist, index=1):
        for reporef in rrlist:
            if not reporef or not reporef.ref:
                continue

            if reporef.get_privacy():
                priv_text = ' priv="1"'
            else:
                priv_text = ""

            if reporef.call_number == "":
                callno_text = ""
            else:
                callno_text = ' callno="%s"' % escxml(reporef.call_number)

            mtype = reporef.media_type.xml_str()
            if mtype:
                type_text = ' medium="%s"' % escxml(mtype)
            else:
                type_text = ""

            note_list = reporef.get_note_list()
            if len(note_list) == 0:
                self.write_ref(
                    "reporef",
                    reporef.ref,
                    index,
                    close=True,
                    extra_text=priv_text + callno_text + type_text,
                )
            else:
                self.write_ref(
                    "reporef",
                    reporef.ref,
                    index,
                    close=False,
                    extra_text=priv_text + callno_text + type_text,
                )
                self.write_note_list(note_list, index + 1)
                sp = "  " * index
                self.g.write("%s</reporef>\n" % sp)

    def write_url_list(self, list, index=1):
        sp = "  " * index
        for url in list:
            url_type = url.get_type().xml_str()
            if url_type:
                type_text = ' type="%s"' % escxml(url_type)
            else:
                type_text = ""
            priv_text = conf_priv(url)
            if url.get_description() != "":
                desc_text = ' description="%s"' % self.fix(url.get_description())
            else:
                desc_text = ""
            path_text = '  href="%s"' % self.fix(url.get_path())
            self.g.write(
                "%s<url%s%s%s%s/>\n" % (sp, priv_text, path_text, type_text, desc_text)
            )

    def write_place_obj(self, place, index=1):
        self.write_primary_tag("placeobj", place, index, close=False)
        ptype = self.fix(place.get_type().xml_str())
        self.g.write(' type="%s"' % ptype)
        self.g.write(">\n")

        title = self.fix(place.get_title())
        code = self.fix(place.get_code())
        self.write_line_nofix("ptitle", title, index + 1)
        self.write_line_nofix("code", code, index + 1)

        self.dump_place_name(place.get_name(), index + 1)
        for pname in place.get_alternative_names():
            self.dump_place_name(pname, index + 1)

        longitude = self.fix(place.get_longitude())
        lat = self.fix(place.get_latitude())
        if longitude or lat:
            self.g.write(
                '%s<coord long="%s" lat="%s"/>\n' % ("  " * (index + 1), longitude, lat)
            )
        for placeref in place.get_placeref_list():
            self.dump_place_ref(placeref, index + 1)
        list(map(self.dump_location, place.get_alternate_locations()))
        self.write_media_list(place.get_media_list(), index + 1)
        self.write_url_list(place.get_url_list())
        self.write_note_list(place.get_note_list(), index + 1)
        for citation_handle in place.get_citation_list():
            self.write_ref("citationref", citation_handle, index + 1)

        for tag_handle in place.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</placeobj>\n" % ("  " * index))

    def write_object(self, obj, index=1):
        self.write_primary_tag("object", obj, index)
        handle = obj.get_gramps_id()
        mime_type = obj.get_mime_type()
        path = obj.get_path()
        desc = obj.get_description()
        if desc:
            desc_text = ' description="%s"' % self.fix(desc)
        else:
            desc_text = ""
        checksum = obj.get_checksum()
        if checksum:
            checksum_text = ' checksum="%s"' % checksum
        else:
            checksum_text = ""
        if self.strip_photos == 1:
            path = os.path.basename(path)
        elif self.strip_photos == 2 and (len(path) > 0 and os.path.isabs(path)):
            drive, path = os.path.splitdrive(path)
            path = path[1:]
        if win():
            # Always export path with \ replaced with /. Otherwise import
            # from Windows to Linux of gpkg's path to images does not work.
            path = path.replace("\\", "/")
        self.g.write(
            '%s<file src="%s" mime="%s"%s%s/>\n'
            % (
                "  " * (index + 1),
                self.fix(path),
                self.fix(mime_type),
                checksum_text,
                desc_text,
            )
        )
        self.write_attribute_list(obj.get_attribute_list())
        self.write_note_list(obj.get_note_list(), index + 1)
        dval = obj.get_date_object()
        if not dval.is_empty():
            self.write_date(dval, index + 1)
        for citation_handle in obj.get_citation_list():
            self.write_ref("citationref", citation_handle, index + 1)

        for tag_handle in obj.get_tag_list():
            self.write_ref("tagref", tag_handle, index + 1)

        self.g.write("%s</object>\n" % ("  " * index))


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def sortById(first, second):
    fid = first.get_gramps_id()
    sid = second.get_gramps_id()

    if fid < sid:
        return -1
    else:
        return fid != sid


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def conf_priv(obj):
    if obj.get_privacy() != 0:
        return ' priv="%d"' % obj.get_privacy()
    else:
        return ""


# -------------------------------------------------------------------------
#
# export_data
#
# -------------------------------------------------------------------------
def export_data(database, filename, user, option_box=None):
    """
    Call the XML writer with the syntax expected by the export plugin.
    """
    if os.path.isfile(filename):
        try:
            shutil.copyfile(filename, filename + ".bak")
            shutil.copystat(filename, filename + ".bak")
        except:
            pass

    compress = _gzip_ok == 1

    if option_box:
        option_box.parse_options()
        database = option_box.get_filtered_database(database)
        compress = compress and option_box.get_use_compression()

    g = XmlWriter(database, user, 0, compress)
    return g.write(filename)


# -------------------------------------------------------------------------
#
# XmlWriter
#
# -------------------------------------------------------------------------
class XmlWriter(GrampsXmlWriter):
    """
    Writes a database to the XML file.
    """

    def __init__(self, dbase, user, strip_photos, compress=1):
        GrampsXmlWriter.__init__(self, dbase, strip_photos, compress, VERSION, user)
        self.user = user

    def write(self, filename):
        """
        Write the database to the specified file.
        """
        ret = 0  # False
        try:
            ret = GrampsXmlWriter.write(self, filename)
        except DbWriteFailure as msg:
            (m1, m2) = msg.messages()
            self.user.notify_error("%s\n%s" % (m1, m2))
        return ret
